/* {{{ Copyright Notices */

/*
 * Copyright 2006 - Wind River Systems
 *
 * lttdaemon -
 *
 * Extended version of lttd with bits taken from lttd.c liblttcl.c and
 * lttctl.c and the old tracedaemon.c to satisfy the requirements of the
 * Wind River System Viewer and the Wind River Linux distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* lttd
 *
 * Linux Trace Toolkit Daemon
 *
 * This is a simple daemon that reads a few relayfs channels and save them in a
 * trace.
 *
 *
 * Copyright 2005 -
 *	Mathieu Desnoyers <mathieu.desnoyers@polymtl.ca>
 */

/* libltt
 *
 * Linux Trace Toolkit Netlink Control Library
 *
 * Controls the ltt-control kernel module through a netlink socket.
 *
 * Heavily inspired from libipq.c (iptables) made by
 * James Morris <jmorris@intercode.com.au>
 *
 * Copyright 2005 -
 *	Mathieu Desnoyers <mathieu.desnoyers@polymtl.ca>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* lttctl
 *
 * Linux Trace Toolkit Control
 *
 * Small program that controls LTT through libltt.
 *
 * Copyright 2005 -
 *	Mathieu Desnoyers <mathieu.desnoyers@polymtl.ca>
 */

/*
    TraceDaemon.c : Tracing daemon for the Linux kernel.
    Copyright (C) 1999, 2000, 2001, 2002 Karim Yaghmour (karym@opersys.com).

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* }}} */

/* {{{ includes */

#define _GNU_SOURCE		/* asprintf, etc */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/mman.h>
#include <sys/mount.h>

/* lttctl/netlink */
#include <asm/types.h>
#include <sys/socket.h>
#include <linux/netlink.h>

/* }}} */

/* {{{ ftw.h */

/* Copyright (C) 1992,1996,1997,1998,1999,2003 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

/*
 *	X/Open Portability Guide 4.2: ftw.h
 */

#ifndef _FTW_H
#define	_FTW_H	1

#include <features.h>

#include <sys/types.h>
#include <sys/stat.h>


__BEGIN_DECLS

/* Values for the FLAG argument to the user function passed to `ftw'
   and 'nftw'.  */
enum
{
  FTW_F,		/* Regular file.  */
#define FTW_F	 FTW_F
  FTW_D,		/* Directory.  */
#define FTW_D	 FTW_D
  FTW_DNR,		/* Unreadable directory.  */
#define FTW_DNR	 FTW_DNR
  FTW_NS,		/* Unstatable file.  */
#define FTW_NS	 FTW_NS

#if defined __USE_BSD || defined __USE_XOPEN_EXTENDED

  FTW_SL,		/* Symbolic link.  */
# define FTW_SL	 FTW_SL
#endif

#ifdef __USE_XOPEN_EXTENDED
/* These flags are only passed from the `nftw' function.  */
  FTW_DP,		/* Directory, all subdirs have been visited. */
# define FTW_DP	 FTW_DP
  FTW_SLN		/* Symbolic link naming non-existing file.  */
# define FTW_SLN FTW_SLN

#endif	/* extended X/Open */
};


#ifdef __USE_XOPEN_EXTENDED
/* Flags for fourth argument of `nftw'.  */
enum
{
  FTW_PHYS = 1,		/* Perform physical walk, ignore symlinks.  */
# define FTW_PHYS	FTW_PHYS
  FTW_MOUNT = 2,	/* Report only files on same file system as the
			   argument.  */
# define FTW_MOUNT	FTW_MOUNT
  FTW_CHDIR = 4,	/* Change to current directory while processing it.  */
# define FTW_CHDIR	FTW_CHDIR
  FTW_DEPTH = 8		/* Report files in directory before directory itself.*/
# define FTW_DEPTH	FTW_DEPTH
};

/* Structure used for fourth argument to callback function for `nftw'.  */
struct FTW
  {
    int base;
    int level;
  };
#endif	/* extended X/Open */


/* Convenient types for callback functions.  */
typedef int (*__ftw_func_t) (__const char *__filename,
			     __const struct stat *__status, int __flag);
#ifdef __USE_LARGEFILE64
typedef int (*__ftw64_func_t) (__const char *__filename,
			       __const struct stat64 *__status, int __flag);
#endif
#ifdef __USE_XOPEN_EXTENDED
typedef int (*__nftw_func_t) (__const char *__filename,
			      __const struct stat *__status, int __flag,
			      struct FTW *__info);
# ifdef __USE_LARGEFILE64
typedef int (*__nftw64_func_t) (__const char *__filename,
				__const struct stat64 *__status,
				int __flag, struct FTW *__info);
# endif
#endif

/* Call a function on every element in a directory tree.

   This function is a possible cancellation point and therefore not
   marked with __THROW.  */
#ifndef __USE_FILE_OFFSET64
extern int ftw (__const char *__dir, __ftw_func_t __func, int __descriptors);
#else
# ifdef __REDIRECT
extern int __REDIRECT (ftw, (__const char *__dir, __ftw_func_t __func,
			     int __descriptors), ftw64);
# else
#  define ftw ftw64
# endif
#endif
#ifdef __USE_LARGEFILE64
extern int ftw64 (__const char *__dir, __ftw64_func_t __func,
		  int __descriptors);
#endif

#ifdef __USE_XOPEN_EXTENDED
/* Call a function on every element in a directory tree.  FLAG allows
   to specify the behaviour more detailed.

   This function is a possible cancellation point and therefore not
   marked with __THROW.  */
# ifndef __USE_FILE_OFFSET64
extern int nftw (__const char *__dir, __nftw_func_t __func, int __descriptors,
		 int __flag);
# else
#  ifdef __REDIRECT
extern int __REDIRECT (nftw, (__const char *__dir, __nftw_func_t __func,
			      int __descriptors, int __flag), nftw64);
#  else
#   define nftw nftw64
#  endif
# endif
# ifdef __USE_LARGEFILE64
extern int nftw64 (__const char *__dir, __nftw64_func_t __func,
		   int __descriptors, int __flag);
# endif
#endif

__END_DECLS

#endif	/* ftw.h */

/* }}} */

/* {{{ #defines */

#define OK	0
#define ERROR	1

#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

/*
 * Exit status values:
 *
 * EXIT_FAILURE is 1, so we differentiate by starting above that.
 */
#define EXIT_STATUS_WITH_WARNINGS		2
#define EXIT_STATUS_CANNOT_OPEN_LOGGING_STREAMS 126
#define EXIT_STATUS_OUT_OF_MEMORY		127

#define MAX_TRACE_FILES 255

#define LOG_MESSAGES_WITH_TIME

#define LOG_MSG(format, ...) { \
    log_msg (stdout, format, ## __VA_ARGS__); \
    log_msg (lttdaemon_logstream, format, ## __VA_ARGS__); }

#define LOG_ERR(format, ...) { \
    log_err (stderr, format, ## __VA_ARGS__); \
    log_err (lttdaemon_logstream, format, ## __VA_ARGS__); }

#define LOG_WARN(format, ...) { \
    log_warn (stderr, format, ## __VA_ARGS__); \
    log_warn (lttdaemon_logstream, format, ## __VA_ARGS__); }

/* }}} */

/* {{{ Function Prototypes */

static void
die (int exit_status, const char *fmt, ...)
     __attribute__ ((format (printf, 2, 3)));

static void
log_err (FILE *stream, const char *fmt, ...)
     __attribute__ ((format (printf, 2, 3)));

static void
log_msg (FILE *stream, const char *fmt, ...)
     __attribute__ ((format (printf, 2, 3)));

static void
log_warn (FILE *stream, const char *fmt, ...)
     __attribute__ ((format (printf, 2, 3)));

static char*
format_msg (const char *fmt, ...)
     __attribute__ ((format (printf, 1, 2)));

static void
lttdaemon_exit (int exit_status) __attribute__ ((noreturn));

static void *
lttdaemon_malloc (size_t size);

/* }}} */

/* {{{ Typedefs */

typedef enum
  {
    STARTED,
    STOPPED,
    NOTSTARTED
  } logging_status_t;

typedef struct _channel_trace_pair
{
  int channel_fd;
  int trace_fd;
  void *channel_mmap_addr;
  int hungup;
  char *channel_name;
  char *trace_name;
  off_t trace_name_file_size;
  int trace_fd_write_error;
} channel_trace_pair_t;

/* }}} */

/* {{{ Global variables */

static const char wrlinux_version[] = "wrlinux-1.4";

static logging_status_t logging_status = NOTSTARTED;

static int channel_trace_pair_count = 0;
static channel_trace_pair_t channel_trace_pair[MAX_TRACE_FILES];

/* Signal receipt. */
static int have_sigusr1 = 0;
static int have_sigusr2 = 0;
static int have_sighup = 0;
static int have_sigint = 0;
static int have_sigterm = 0;

static const char *progname = "lttdaemon";
static const char *tmp_dir = "/tmp";

/* Log file pathname. */
static char *lttdaemon_logfile = 0;

/* Log file stream. */
static FILE* lttdaemon_logstream = 0;

/* modprobe lttcontrol */
static const char *modprobe_lttcontrol_cmd = "modprobe ltt-control";

/* relayfs */
static const char *relayfs_mountpoint = "/mnt/relay";

/* LTTV */
static const char *ltt_facilities_dir = "/usr/share/ltt-control/facilities";

/* ltt channels */
static const char *ltt_trace_pathname = "/mnt/relay/ltt/trace";

/* Pathnames to various lttdaemon files.  */
static const char *lttdaemon_named_pipe = "/tmp/.lttdaemon.pipe";
static const char *lttdaemon_pidfile = "/tmp/lttdaemon.pid";
static const char *lttdaemon_statusfile = "/tmp/lttdaemon.status";
static const char *lttdaemon_versionfile = "/tmp/lttdaemon.version";

/* Number of errors raised through log_err(). */
static int errors = 0;

/* Number of warnings raised through log_warn(). */
static int warnings = 0;

/* The exit status; used in the atexit() handler. */
static int lttdaemon_exit_status = 0;

/* command line options. */
static int cliopt_append_mode = 0;
static int cliopt_collect_log = 0;
static int cliopt_nsubufs = 2;
static int cliopt_pause = 0;
static int cliopt_resume = 0;
static int cliopt_flight_recorder_mode = 0;
static int cliopt_normal_mode = 0;
static size_t cliopt_subbuf_size = 1024 * 1024;	/* 1MB */
static int max_subbuf_size = 1024 * 1024 * 16;  /* 16MB */
static int max_subbufs = 16;
static int cliopt_verbose = 0;
static int cliopt_query_status = 0;
static char *cliopt_output_dir = 0;
static const char *cliopt_trace_name = "trace";
static int cliopt_load_ltt_control_module = 0;

/* Output file/dir list streams. */
static FILE *file_list_stream = NULL;

/* lttctl */
static struct lttctl_handle * lttctl_handle = NULL;

/* }}} */

/* {{{ liblttctl */

/* Get the next sub buffer that can be read. */
#define RELAYFS_GET_SUBBUF        _IOR(0xF4, 0x00,__u32)

/* Release the oldest reserved (by "get") sub buffer. */
#define RELAYFS_PUT_SUBBUF        _IOW(0xF4, 0x01,__u32)

/* returns the number of sub buffers in the per cpu channel. */
#define RELAYFS_GET_N_SUBBUFS     _IOR(0xF4, 0x02,__u32)

/* returns the size of the sub buffers. */
#define RELAYFS_GET_SUBBUF_SIZE   _IOR(0xF4, 0x03,__u32)

enum LttDaemon
  {
    GET_SUBBUF,
    PUT_SUBBUF,
    GET_N_BUBBUFS,
    GET_SUBBUF_SIZE
  };

#ifndef NETLINK_LTT
#define NETLINK_LTT 31
#endif

#define LTTCTLM_BASE	0x10
#define LTTCTLM_CONTROL	(LTTCTLM_BASE + 1)	/* LTT control message */

enum trace_op
  {
    OP_CREATE,
    OP_DESTROY,
    OP_START,
    OP_STOP,
    OP_NONE
  };

enum trace_mode
  {
    LTT_TRACE_NORMAL,
    LTT_TRACE_FLIGHT
  };

typedef struct lttctl_peer_msg
{
  char trace_name[NAME_MAX];
  enum trace_op op;
  union
  {
    struct
    {
      enum trace_mode mode;
      unsigned subbuf_size;
      unsigned n_subbufs;
    } new_trace;
  } args;
} lttctl_peer_msg_t;

struct lttctl_handle
{
  int fd;
  struct sockaddr_nl local;
  struct sockaddr_nl peer;
};

typedef struct lttctl_resp_msg
{
  int err;
} lttctl_resp_msg_t;

struct lttctl_handle *lttctl_create_handle(void);

enum LttCtlErrors
  {
    LTTCTL_ERR_NONE = 0,
    LTTCTL_ERR_IMPL,
    LTTCTL_ERR_HANDLE,
    LTTCTL_ERR_SOCKET,
    LTTCTL_ERR_BIND,
    LTTCTL_ERR_BUFFER,
    LTTCTL_ERR_RECV,
    LTTCTL_ERR_NLEOF,
    LTTCTL_ERR_ADDRLEN,
    LTTCTL_ERR_STRUNC,
    LTTCTL_ERR_RTRUNC,
    LTTCTL_ERR_NLRECV,
    LTTCTL_ERR_SEND,
    LTTCTL_ERR_SUPP,
    LTTCTL_ERR_RECVBUF,
    LTTCTL_ERR_TIMEOUT,
    LTTCTL_ERR_PROTOCOL
  };

#define LTTCTL_MAXERR LTTCTL_ERR_PROTOCOL

struct lttctl_errmap_t
{
  int errcode;
  const char *message;
} lttctl_errmap[] =
  {
    { LTTCTL_ERR_NONE, "Unknown error" },
    { LTTCTL_ERR_IMPL, "Implementation error" },
    { LTTCTL_ERR_HANDLE, "Unable to create netlink handle" },
    { LTTCTL_ERR_SOCKET, "Unable to create netlink socket" },
    { LTTCTL_ERR_BIND, "Unable to bind netlink socket" },
    { LTTCTL_ERR_BUFFER, "Unable to allocate buffer" },
    { LTTCTL_ERR_RECV, "Failed to receive netlink message" },
    { LTTCTL_ERR_NLEOF, "Received EOF on netlink socket" },
    { LTTCTL_ERR_ADDRLEN, "Invalid peer address length" },
    { LTTCTL_ERR_STRUNC, "Sent message truncated" },
    { LTTCTL_ERR_RTRUNC, "Received message truncated" },
    { LTTCTL_ERR_NLRECV, "Received error from netlink" },
    { LTTCTL_ERR_SEND, "Failed to send netlink message" },
    { LTTCTL_ERR_SUPP, "Operation not supported" },
    { LTTCTL_ERR_RECVBUF, "Receive buffer size invalid" },
    { LTTCTL_ERR_TIMEOUT, "Timeout" },
    { LTTCTL_ERR_PROTOCOL, "Invalid protocol specified" }
  };

static int lttctl_errno = LTTCTL_ERR_NONE;

static const char *
lttctl_strerror(int errcode)
{
  if (errcode < 0 || errcode > LTTCTL_MAXERR)
    errcode = LTTCTL_ERR_IMPL;

  return lttctl_errmap[errcode].message;
}

static const char *
lttctl_errstr (void)
{
  return lttctl_strerror(lttctl_errno);
}

static void
lttctl_perror (const char *s)
{
  if (lttctl_errno)
    LOG_ERR ("%s: %s\n", s, lttctl_errstr ());

  if (errno)
    LOG_ERR ("%s: %s\n", s, strerror(errno));
}

static ssize_t
lttctl_netlink_sendto(const struct lttctl_handle *h,
		      const void *msg,
		      size_t len)
{
  int status = sendto (h->fd, msg, len, 0,
		       (struct sockaddr *)&h->peer, sizeof(h->peer));

  if (status < 0)
    lttctl_errno = LTTCTL_ERR_SEND;

  return status;
}

static ssize_t
lttctl_netlink_recvfrom (const struct lttctl_handle *h,
			 unsigned char *buf,
			 size_t len,
			 int timeout)
{
  int addrlen, status;
  struct nlmsghdr *nlh;

  if (len < sizeof(struct nlmsghdr))
    {
      lttctl_errno = LTTCTL_ERR_RECVBUF;
      lttctl_perror ("Netlink recvfrom");
      return -1;
    }

  addrlen = sizeof (h->peer);

  if (timeout != 0)
    {
      int ret;
      struct timeval tv;
      fd_set read_fds;

      if (timeout < 0)
	{
	  /* non-block non-timeout */
	  tv.tv_sec = 0;
	  tv.tv_usec = 0;
	}
      else
	{
	  tv.tv_sec = timeout / 1000000;
	  tv.tv_usec = timeout % 1000000;
	}

      FD_ZERO(&read_fds);
      FD_SET(h->fd, &read_fds);

    select_again:

      ret = select(h->fd+1, &read_fds, NULL, NULL, &tv);

      if (ret < 0)
	{
	  if (errno == EINTR)
	    {
	      goto select_again;
	      return 0;
	    }
	  else
	    {
	      lttctl_errno = LTTCTL_ERR_RECV;
	      lttctl_perror("Netlink recvfrom");
	      return -1;
	    }
	}

      if (!FD_ISSET(h->fd, &read_fds))
	{
	  lttctl_errno = LTTCTL_ERR_TIMEOUT;
	  return 0;
	}
    }

  status = recvfrom (h->fd, buf, len, 0,
		     (struct sockaddr *)&h->peer, &addrlen);

  if (status < 0)
    {
      lttctl_errno = LTTCTL_ERR_RECV;
      lttctl_perror("Netlink recvfrom");
      return status;
    }

  if (addrlen != sizeof(h->peer))
    {
      lttctl_errno = LTTCTL_ERR_RECV;
      lttctl_perror("Netlink recvfrom");
      return -1;
    }

  if (h->peer.nl_pid != 0)
    {
      lttctl_errno = LTTCTL_ERR_RECV;
      lttctl_perror("Netlink recvfrom");
      return -1;
    }

  if (status == 0)
    {
      lttctl_errno = LTTCTL_ERR_NLEOF;
      lttctl_perror("Netlink recvfrom");
      return -1;
    }

  nlh = (struct nlmsghdr *)buf;

  if (nlh->nlmsg_flags & MSG_TRUNC || nlh->nlmsg_len > (unsigned) status)
    {
      lttctl_errno = LTTCTL_ERR_RTRUNC;
      lttctl_perror("Netlink recvfrom");
      return -1;
    }

  return status;
}

/* public interface */

/*
 * Create and initialise an lttctl handle.
 */
struct lttctl_handle *
lttctl_create_handle(void)
{
  int status;
  struct lttctl_handle *h;

  h = (struct lttctl_handle *)
    lttdaemon_malloc (sizeof(struct lttctl_handle));

  if (h == NULL)
    {
      lttctl_errno = LTTCTL_ERR_HANDLE;
      lttctl_perror("Create handle");
      goto alloc_error;
    }

  memset(h, 0, sizeof(struct lttctl_handle));

  h->fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_LTT);

  if (h->fd == -1)
    {
      lttctl_errno = LTTCTL_ERR_SOCKET;
      lttctl_perror("Create handle");
      goto socket_error;
    }

  memset(&h->local, 0, sizeof(struct sockaddr_nl));
  h->local.nl_family = AF_NETLINK;
  h->local.nl_pid = getpid();
  h->local.nl_groups = 0;

  status = bind(h->fd, (struct sockaddr *)&h->local, sizeof(h->local));

  if (status == -1)
    {
      lttctl_errno = LTTCTL_ERR_BIND;
      lttctl_perror("Create handle");
      goto bind_error;
    }

  memset(&h->peer, 0, sizeof(struct sockaddr_nl));
  h->peer.nl_family = AF_NETLINK;
  h->peer.nl_pid = 0;
  h->peer.nl_groups = 0;

  return h;

  /* Error condition */
 bind_error:
 socket_error:
  close(h->fd);
 alloc_error:
  free(h);
  return NULL;
}

/*
 * No error condition is checked here at this stage, but it may happen
 * if/when reliable messaging is implemented.
 */
static int
lttctl_destroy_handle(struct lttctl_handle *h)
{
  if (h)
    {
      close(h->fd);
      free(h);
    }

  return 0;
}

static int
lttctl_create_trace (const struct lttctl_handle *h,
		     const char *name,
		     enum trace_mode mode,
		     unsigned subbuf_size,
		     unsigned n_subbufs)
{
  int err;

  struct
  {
    struct nlmsghdr	nlh;
    lttctl_peer_msg_t	msg;
  } req;

  struct
  {
    struct nlmsghdr	nlh;
    struct nlmsgerr	nlerr;
    lttctl_peer_msg_t	msg;
  } ack;

  memset(&req, 0, sizeof(req));
  req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(lttctl_peer_msg_t));
  req.nlh.nlmsg_flags = NLM_F_REQUEST|NLM_F_ACK;
  req.nlh.nlmsg_type = LTTCTLM_CONTROL;
  req.nlh.nlmsg_pid = h->local.nl_pid;
  req.nlh.nlmsg_seq = 0;

  strncpy(req.msg.trace_name, name, NAME_MAX);
  req.msg.op = OP_CREATE;
  req.msg.args.new_trace.mode = mode;
  req.msg.args.new_trace.subbuf_size = subbuf_size;
  req.msg.args.new_trace.n_subbufs = n_subbufs;

  err = lttctl_netlink_sendto(h, (void *)&req, req.nlh.nlmsg_len);

  if (err < 0)
    goto senderr;

  err = lttctl_netlink_recvfrom(h, (void*)&ack, sizeof(ack), 0);

  if (err < 0)
    goto senderr;

  err = ack.nlerr.error;

  if (err != 0)
    {
      errno = err;
      lttctl_perror("Create Trace Error");
      return err;
    }

  return 0;

 senderr:
  lttctl_perror("Create Trace Error");
  err = EPERM;
  return err;
}

static int
lttctl_destroy_trace (const struct lttctl_handle *h,
		      const char *name)
{
  struct
  {
    struct nlmsghdr	nlh;
    lttctl_peer_msg_t	msg;
  } req;

  struct
  {
    struct nlmsghdr	nlh;
    struct nlmsgerr	nlerr;
    lttctl_peer_msg_t	msg;
  } ack;
  int err;

  memset(&req, 0, sizeof(req));
  req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(lttctl_peer_msg_t));
  req.nlh.nlmsg_flags = NLM_F_REQUEST;
  req.nlh.nlmsg_type = LTTCTLM_CONTROL;
  req.nlh.nlmsg_pid = h->local.nl_pid;

  strncpy(req.msg.trace_name, name, NAME_MAX);
  req.msg.op = OP_DESTROY;

  err = lttctl_netlink_sendto(h, (void *)&req, req.nlh.nlmsg_len);

  if (err < 0)
    goto senderr;

  err = lttctl_netlink_recvfrom(h, (void*)&ack, sizeof(ack), 0);

  if (err < 0)
    goto senderr;

  err = ack.nlerr.error;

  if (err != 0)
    {
      errno = err;
      lttctl_perror("Destroy Trace Channels Error");
      return err;
    }

  return 0;

 senderr:
  lttctl_perror("Destroy Trace Channels Error");
  err = EPERM;
  return err;
}

static int
lttctl_start (const struct lttctl_handle *h,
	      const char *name)
{
  struct
  {
    struct nlmsghdr	nlh;
    lttctl_peer_msg_t	msg;
  } req;

  struct
  {
    struct nlmsghdr	nlh;
    struct nlmsgerr	nlerr;
    lttctl_peer_msg_t	msg;
  } ack;

  int err;

  memset(&req, 0, sizeof(req));
  req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(lttctl_peer_msg_t));
  req.nlh.nlmsg_flags = NLM_F_REQUEST;
  req.nlh.nlmsg_type = LTTCTLM_CONTROL;
  req.nlh.nlmsg_pid = h->local.nl_pid;

  strncpy(req.msg.trace_name, name, NAME_MAX);
  req.msg.op = OP_START;

  err = lttctl_netlink_sendto(h, (void *)&req, req.nlh.nlmsg_len);

  if (err < 0)
    goto senderr;

  err = lttctl_netlink_recvfrom(h, (void*)&ack, sizeof(ack), 0);

  if (err < 0)
    goto senderr;

  err = ack.nlerr.error;

  if (err != 0)
    {
      errno = err;
      LOG_ERR ("start trace: %s: %s\n", ltt_trace_pathname, strerror (errno));
      /*       lttctl_perror("Start Trace Error"); */
      return err;
    }

  logging_status = STARTED;

  return 0;

 senderr:
  err = EPERM;
  LOG_ERR ("start trace: %s: %s\n", ltt_trace_pathname, strerror (errno));
  lttctl_perror("Start Trace Error");
  return err;
}

static int
lttctl_stop (const struct lttctl_handle *h,
	     const char *name)
{
  struct
  {
    struct nlmsghdr	nlh;
    lttctl_peer_msg_t	msg;
  } req;
  struct
  {
    struct nlmsghdr	nlh;
    struct nlmsgerr	nlerr;
    lttctl_peer_msg_t	msg;
  } ack;
  int err;

  memset(&req, 0, sizeof(req));
  req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(lttctl_peer_msg_t));
  req.nlh.nlmsg_flags = NLM_F_REQUEST;
  req.nlh.nlmsg_type = LTTCTLM_CONTROL;
  req.nlh.nlmsg_pid = h->local.nl_pid;

  strncpy(req.msg.trace_name, name, NAME_MAX);
  req.msg.op = OP_STOP;

  err = lttctl_netlink_sendto(h, (void *)&req, req.nlh.nlmsg_len);

  if (err < 0)
    goto senderr;

  err = lttctl_netlink_recvfrom(h, (void*)&ack, sizeof(ack), 0);

  if (err < 0)
      goto senderr;

  err = ack.nlerr.error;

  if (err != 0)
    {
      errno = err;
      LOG_ERR ("stop trace: %s: %s\n", ltt_trace_pathname, strerror (errno));
      return err;
    }

  logging_status = STOPPED;

  return 0;

 senderr:
  err = EPERM;
  lttctl_perror("Stop Trace Error");
  return err;
}

/* }}} */

/* {{{ MapSlashProc - from old tracedaemon */

#define TD_FILE_NAME_STRLEN       40  /* File name string length */
#define TD_STATUS_STRLEN         500  /* String length to read from process status file */
#define TD_PROCES_NAME_STRLEN    256  /* Length of process name */
#define MAX_CPUS                  32  /* Same as kernel NR_CPUS */

/*******************************************************************
 * Function :
 *    MapProc()
 * Description :
 *    Function maping out /proc.
 * Parameters :
 *    pmProcFile, file where the information is to be written
 * Return values :
 *    NONE
 * History :
 *    K. Yaghmour, 05/99, Initial typing.
 *******************************************************************/
static void
MapProc(char *pmProcFileName)
{
  int             i;                               /* Generic index */
  int             lIRQ;                            /* IRQ ID */
  int             lPID;                            /* Process' PID */
  int             lTaskPID;                        /* Process' Task PID */
  int             lPPID;                           /* Parent's PID */
  int             lNbCPU;                          /* Number of CPUs */
  int             lSizeRead;                       /* Quantity of data read from file */
  int             lStatusFile;                     /* File containd process status */
  char            lFileName[TD_FILE_NAME_STRLEN];  /* A string containing a complete file name */
  char            lDataBuf[TD_STATUS_STRLEN];      /* Status file content */
  char            lName[TD_PROCES_NAME_STRLEN];    /* Name of the process */
  char*           lString;                         /* Generic string */
  DIR*            pDir;                            /* Directory pointer */
  DIR*            pTaskDir;                        /* Directory pointer */
  FILE*           pIntFile;                        /* Int pile pointer */
  struct dirent*  pDirEntry;                       /* Directory entry */
  struct dirent*  pTaskDirEntry;                   /* Directory entry */

  char            typename [TD_STATUS_STRLEN];     /* typename of interrupt */
  int             numChars = 0;                    /* Number of chars read during scanf */
  int             numInts;                         /* 'count' field from /proc/interrupts */
  int             numFields;                       /* Result of sscanf() */
  int             pmProcFile;

  /* Map out /proc into user given file */

  /* Open /proc */
  if((pDir = opendir("/proc")) == NULL)
    {
      LOG_WARN ("Your system doesn't support /proc\n");
      return;
    }

  if (cliopt_append_mode)
    {
      pmProcFile = open(pmProcFileName, O_RDWR | O_CREAT | O_APPEND, 0666);
    }
  else
    {
      pmProcFile = open(pmProcFileName, O_RDWR | O_CREAT | O_TRUNC, 0666);
    }

  if (pmProcFile < 0)
    {
      /* File ain't good */
      LOG_WARN ("Unable to open output file (%s) to dump current process information: %s\n",
		pmProcFileName ? pmProcFileName : "(null)",
		strerror (errno));
      return;
    }

  /* Read the directory through */
  while((pDirEntry = readdir(pDir)) != NULL)
    {
      /* Is this a directory entry describing a PID */
      if(isdigit(pDirEntry->d_name[0]))
	{
	  /* Get the PID in number form */
	  lPID = atoi(pDirEntry->d_name);

	  /* Open the status file */
	  snprintf(lFileName, TD_FILE_NAME_STRLEN, "/proc/%d/status", lPID);
	  if((lStatusFile = open(lFileName, O_RDONLY, 0)) < 0)
	    continue;

	  /* Read a little bit */
	  if((lSizeRead = read(lStatusFile, lDataBuf, TD_STATUS_STRLEN)) == 0)
	    continue;

	  /* Close the file */
	  close(lStatusFile);

	  /* Find Name and Parent PID */
	  lString  = strstr(lDataBuf, "Name");
	  sscanf(lString, "Name: %s", lName);
	  lString  = strstr(lDataBuf, "PPid");
	  sscanf(lString, "PPid: %d", &lPPID);

	  /* Print out the precious information */
	  snprintf(lDataBuf, TD_STATUS_STRLEN, "PID: %d; PPID: %d; NAME: %s\n", lPID, lPPID, lName);
	  write(pmProcFile, lDataBuf, strlen(lDataBuf));

	  /* Emit thread information, if it exists. */

	  snprintf (lFileName, TD_FILE_NAME_STRLEN, "/proc/%d/task", lPID);

	  /* Open /proc/<pid>/task */

	  if((pTaskDir = opendir(lFileName)) == NULL)
	    continue;

	  /* Read the directory through */
	  while((pTaskDirEntry = readdir(pTaskDir)) != NULL)
	    {
	      /* Is this a directory entry describing a PID */
	      if(isdigit(pTaskDirEntry->d_name[0]))
		{
		  /* Get the PID in number form */
		  lTaskPID = atoi(pTaskDirEntry->d_name);

		  /* the task directory always contains an entry for the
		     current pid, but as we've just written that information
		     we just skip this instance. */

		  if (lTaskPID == lPID)
		    continue;

		  snprintf(lDataBuf, TD_STATUS_STRLEN,
			   "PID: %d; PPID: %d; NAME: %s\n",
			   lTaskPID, lPID, lName);

		  write(pmProcFile, lDataBuf, strlen(lDataBuf));
		}
	    }

	  /* Close the task directory */
	  closedir(pTaskDir);
	}
    }

  /* Close the directory */
  closedir(pDir);

  /* Open the interrupts information file */
  if((pIntFile = fopen("/proc/interrupts", "r")) == NULL)
    {
      LOG_WARN("open /proc/interrupts failed");
      return;
    }

  /* The first line is the column header, usually just CPU0, CPU1, etc.
     Count the number of CPUs based on "CPUn" occurences.  I didn't
     see anything else in /proc that would have been any easier
     and I need to be in this pseudofile anyway... */
  if(fgets(lDataBuf, 200, pIntFile) == NULL)
    { /* Error or EOF */
      LOG_WARN("read /proc/interrupts failed");
      fclose(pIntFile);
      return;
    }

  /* Initialize the number of CPUs */
  lNbCPU = 0;

  /* Print header for CPU 0 */
  snprintf(lName, TD_PROCES_NAME_STRLEN, "CPU%d", lNbCPU);

  /* Find entry for CPU 0 */
  lString = strstr(lDataBuf, lName);

  /* Go through rest of string */
  while (lString)
    {
      /* Look for next CPU info */
      lNbCPU++;
      snprintf(lName, TD_PROCES_NAME_STRLEN, "CPU%d", lNbCPU);
      lString = strstr(lString, lName);
    }

  /* How many CPUs did we find */
  if (lNbCPU <= 0 || lNbCPU > 128)
    {
      /* Missing or too weird */
      LOG_WARN("Can't determine CPU count /proc/interrupts\n");
      fclose(pIntFile);
      return;
    }

  /* Read a line from file and parse it. There will be an interrupt
     name, a count for each CPU, and a name. Most architectures also
     have a typename, but on ARM (at least) this is missing. If we don't
     get an interrupt name, we probably got it but thought it was the
     typename.  The presence of white space in the names means the
     results will vary, but should still be reasonable.  Outer fscanf
     stops on EOF, "NMI", "ERR", or other non-numeric chars. */

  while (fgets (lDataBuf, TD_STATUS_STRLEN, pIntFile) != NULL)
    {
      lString = lDataBuf;

      i = strlen (lDataBuf);

      /* Remove EOL */
      if ((i > 0) && (lDataBuf [i - 1] == '\n'))
	lDataBuf [i - 1] = 0;

      /* Read IRQ number (Assume %n does not change the number of
	 conversions) */
      numFields = sscanf (lString, "%d: %n", &lIRQ, &numChars);
      if (numFields != 1)
	break;

      /* Move the startpoint on */
      lString += numChars;

      /* Is this IRQ number valid?  */
      if(lIRQ > 256)
	{
	  LOG_WARN("IRQ %d is out of range in /proc/interrupts\n", lIRQ);
	  fclose(pIntFile);
	  return;
	}

      /* Get int count for each CPU */
      for(i = 0; i < lNbCPU; i++)
	{
	  numFields = sscanf (lString, "%d %n", &numInts, &numChars);
	  if (numFields != 1)
	    {
	      LOG_WARN("Bad INT count field seen in /proc/interrupts\n");
	      fclose(pIntFile);
	      return;
	    }
	  lString += numChars;
	}

      /* Get the typename */
      numFields = sscanf (lString, "%s%n", typename, &numChars);
      if (numFields != 1)
	break;


      /* Treat the remainder of the line as the interrupt name. If there
	 is nothing left, use the typename instead. ARM (maybe others?)
	 omits the typename, so what we see as typename should be the
	 interrupt name */

      strcpy (lName, lString + numChars);
      if (strlen (lName) == 0)
	strcpy (lName, typename);

      /* Format string describing the IRQ */
      snprintf(lDataBuf, TD_STATUS_STRLEN, "IRQ: %d; NAME: %s\n", lIRQ, lName);

      /* Write string to proc file */
      write(pmProcFile, lDataBuf, strlen(lDataBuf));
    }
  /* Close the interrupt information file */
  fclose(pIntFile);

  close (pmProcFile);
}

static void
MapSlashProc (void)
{
  const char *trace_root = cliopt_output_dir;

  char proc_path[PATH_MAX];
  size_t trace_root_len = strlen (trace_root);

  strncpy(proc_path, trace_root, PATH_MAX);
  strncat(proc_path, "/wrsv.proc", PATH_MAX - trace_root_len);

  MapProc (proc_path);
}

/* }}} */

/* {{{ lttdaemon */


/**
 * Returns the current time in ascii format using ctime(); the \n is
 * replaced with \0.  Returns "" on error.
 */
#ifdef LOG_MESSAGES_WITH_TIME
static char*
current_time (void)
{
  time_t now = time (NULL);
  char *s = ctime (&now);

  if (s)
    s[strlen(s)-1] = '\0';		/* remove \n */
  else
    s = (char*) "";

  return s;
}
#endif


/**
 * Logs message to stream annotated with current time and progname:pid.
 */
static void
vflog (FILE* stream,
       const char *type,
       const char *fmt,
       va_list ap)
{
#ifdef LOG_MESSAGES_WITH_TIME
  char *t = current_time ();
#else
  char *t = NULL;
#endif

  /* Print ident stuff first. */

  fprintf (stream, "%s%s[%s:%d]: %s%s",
	   t ? t : "",
	   t ? " " : "",
	   progname,
	   getpid (),
	   (type != NULL) ? type : "",
	   (type != NULL) ? ": " : "");

  vfprintf (stream, fmt, ap);
  fflush (stream);
}

/*
 * Log a message in the same manner as printf() to the message logging
 * stream.
 */
static void
log_msg (FILE *stream, const char *fmt, ...)
{
  if (cliopt_verbose)
    {
      va_list ap;
      va_start (ap, fmt);
      vflog (stream, NULL, fmt, ap);
      va_end (ap);
    }
}


/*
 * Log an error message in the same manner as printf() to the error
 * logging stream.
 */
static void
log_err (FILE *stream, const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  errors++;
  vflog (stream, "error", fmt, ap);
  va_end (ap);
}


/*
 * Log an warning message in the same manner as printf() to the error
 * logging stream.
 */
static void
log_warn (FILE *stream, const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  warnings++;
  vflog (stream, "warning", fmt, ap);
  va_end (ap);
}


/*
 * Exit with 'exit_status' and also record the exit value by logging its
 * value to the error stream.
 */
static void
lttdaemon_exit (int exit_status)
{
  /*
   * Update lttdaemon_exit_status because it is used in the atexit()
   * handler.
   */
  lttdaemon_exit_status = exit_status;

  /* Report exit status values for debugging purposes. */
  LOG_MSG ("exit (%d)\n", exit_status);

  exit (exit_status);
}


/*
 * Prints/formats a message ala printf then calls:
 * lttdaemon_exit(EXIT_FAILURE).
 */
static void
die (int exit_status, const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);

  if (fmt)
    {
      log_err (stderr, fmt, ap);
      log_err (lttdaemon_logstream, fmt, ap);
    }

  va_end (ap);

  lttdaemon_exit (exit_status);
}


/*
 * A wrapper around malloc that automatically die()'s when ENOMEM is
 * raised.
 */
static void *
lttdaemon_malloc (size_t size)
{
  void *p = malloc (size);

  if (p == NULL)
    {
      /*
       * We would normally want to use die() here but if we are out of
       * memory there's no guarantee what will happen.  Either way we want
       * to exit with EXIT_STATUS_OUT_OF_MEMORY.
       */
      LOG_ERR ("malloc failed: %s\n", strerror (errno));
      lttdaemon_exit (EXIT_STATUS_OUT_OF_MEMORY);
    }

  return p;
}


/*
 * A wrapper around realloc that automatically die()'s when ENOMEM is
 * raised.
 */
static void *
lttdaemon_realloc(void *ptr, size_t size)
{
  void *p = realloc (ptr, size);

  if (p == NULL)
    {
      /*
       * We would normally want to use die() here but if we are out of
       * memory there's no guarantee what will happen.  Either way we want
       * to exit with EXIT_STATUS_OUT_OF_MEMORY.
       */

      LOG_ERR ("realloc failed: %s\n", strerror (errno));
      lttdaemon_exit (EXIT_STATUS_OUT_OF_MEMORY);
    }

  return p;
}


/*
 * A wrapper around strdup that automatically die()'s when ENOMEM is
 * raised.
 */
static void *
lttdaemon_strdup (const char *s)
{
  char *p = strdup (s);

  if (p == NULL)
    {
      /*
       * We would normally want to use die() here but if we are out of
       * memory there's no guarantee what will happen.  Either way we want
       * to exit with EXIT_STATUS_OUT_OF_MEMORY.
       */

      LOG_ERR ("strdup failed: %s\n", strerror (errno));
      lttdaemon_exit (EXIT_STATUS_OUT_OF_MEMORY);
    }

  return p;
}


/*
 * sprintf fmt/args into a buffer that grows to accomodate the result.
 * Returns NULL if there was insufficient memory, or a pointer to the
 * formatted message on success.
 */
static char *
format_msg (const char *fmt, ...)
{
  /* Guess we need no more than 100 bytes. */
  int n;
  size_t size = 100;
  char *p;
  va_list ap;

  if ((p = lttdaemon_malloc (size)) == NULL)
    return NULL;

  while (1)
    {
      /* Try to print in the allocated space. */

      va_start(ap, fmt);
      n = vsnprintf (p, size, fmt, ap);
      va_end(ap);

      /* If that worked, return the string. */

      if (n > -1 && n < (int) size)
	return p;

      /* Else try again with more space. */

      if (n > -1)		/* glibc 2.1 */
	size = n+1;		/* precisely what is needed */
      else			/* glibc 2.0 */
	size *= 2;		/* twice the old size */

      if ((p = lttdaemon_realloc (p, size)) == NULL)
	return NULL;
    }
}


/* Make all the directories leading up to PATH, then create PATH.  Note
 * that this changes the process's umask; make sure that all paths
 * leading to a return reset it to ORIGINAL_UMASK. Returns 0 on success,
 * -1 on error.
 */
static int
make_path (const char *path, mode_t nmode, mode_t parent_mode)
{
  mode_t oumask;
  struct stat sb;
  char *p;
  char *npath;

  if (stat (path, &sb) == 0)
    {
      if (S_ISDIR (sb.st_mode) == 0)
	{
	  LOG_ERR ("`%s': file exists but is not a directory\n", path);
	  return -1;
	}

      if (chmod (path, nmode))
	{
	  LOG_ERR ("cannot chmod %s: %s\n", path, strerror (errno));
	  return -1;
	}

      return 0;
    }

  oumask = umask (0);
  npath = lttdaemon_strdup (path); /* So we can write to it. */

  /* Check whether or not we need to do anything with intermediate dirs. */

  p = npath;

  while (*p == '/')		/* Skip leading slashes. */
    p++;

  while ((p = strchr (p, '/')))
    {
      *p = '\0';
      if (stat (npath, &sb) != 0)
	{
	  if (mkdir (npath, parent_mode))
	    {
	      LOG_ERR ("cannot create directory `%s': %s\n",
		       npath,
		       strerror (errno));
	      umask (oumask);
	      free (npath);
	      return -1;
	    }
	}
      else if (S_ISDIR (sb.st_mode) == 0)
	{
	  LOG_ERR ("`%s': file exists but is not a directory\n", npath);
	  umask (oumask);
	  free (npath);
	  return -1;
	}

      *p++ = '/';	/* restore slash */

      while (*p == '/')
	p++;
    }

  /* Create the final directory component. */

  if (stat (npath, &sb) && mkdir (npath, nmode))
    {
      LOG_ERR ("cannot create directory `%s': %s\n", npath, strerror (errno));
      umask (oumask);
      free (npath);
      return -1;
    }

  umask (oumask);
  free (npath);

  return 0;
}

/**
 * make directory PATHNAME and all paths within it.  Returns 0 on
 * success, otherwise non-zero.  All error failures are logged through
 * LOG_ERR().
 */
static int
make_directory_path (const char *pathname)
{
  mode_t nmode;
  mode_t parent_mode;
  mode_t original_umask;
  mode_t omode = S_IRWXU | S_IRWXG | S_IRWXO;	/* a=rwx */

  /* Make the new mode */
  original_umask = umask (0);
  umask (original_umask);

  nmode = (S_IRWXU | S_IRWXG | S_IRWXO) & ~original_umask;
  parent_mode = nmode | (S_IWRITE|S_IEXEC);	/* u+wx */

  /* Adjust new mode based on mode argument */
  nmode &= omode;

  return make_path (pathname, nmode, parent_mode);
}


static int
open_logfile (const char *name)
{
  const char *mode = "w";

  lttdaemon_logfile = format_msg ("%s/%s-%d.log",
				  tmp_dir,
				  name,
				  getpid ());

  assert (lttdaemon_logfile);

  if ((lttdaemon_logstream = fopen (lttdaemon_logfile, mode)) == NULL)
    {
      fprintf (stderr, "cannot open %s: %s\n",
	       lttdaemon_logfile,
	       strerror (errno));
      return EXIT_STATUS_CANNOT_OPEN_LOGGING_STREAMS;
    }

#if 1
  LOG_MSG ("logging to %s\n", lttdaemon_logfile);
#endif

  return OK;
}


/**
 * Convert string 'VAL' to an integer returning OK on success (setting
 * *INTVALUE), otherwise ERROR.  If the string could be converted but
 * would be outside the range of 'INTVALUE' then errno is set to ERANGE.
 */
static int
lttdaemon_atoi (const char *val, int *intvalue)
{
  const char *p = val;
  char *endptr;

  assert (intvalue);

  *intvalue = strtol (p, &endptr, 10);

  if (*p != '\0' && *endptr == '\0')
    return OK;
  else
    return ERROR;
}

/**
 * Iff there were no errors reported, close and delete
 * LTTDAEMON_LOGFILE.
 */
static void
conditionally_close_and_delete_logfile (void)
{
  if (lttdaemon_logstream)
    {
      (void) fclose (lttdaemon_logstream);
      lttdaemon_logstream = NULL;
    }

  if (errors == 0 && warnings == 0 && lttdaemon_exit_status == 0)
    {
      if (lttdaemon_logfile != NULL)
	{
	  unlink (lttdaemon_logfile); /* ignore errors */
	  free (lttdaemon_logfile);
	  lttdaemon_logfile = NULL;
	}
    }
}


/**
 * atexit() handler to delete logging stream.
 */
static void
atexit_handler (void)
{
  conditionally_close_and_delete_logfile ();
}

static int
lttdaemon_stop_destroy_trace (void)
{
  int rc = lttctl_stop (lttctl_handle, cliopt_trace_name);

  if (!rc)
    rc = lttctl_destroy_trace (lttctl_handle, cliopt_trace_name);

  return rc;
}


static int
modprobe_lttcontrol (void)
{
  int rc;
  FILE *fp = 0;
  char buf[4096];			/* line of data from cmd */
  char cmd[4096];
  const char *default_path = "/sbin:/usr/sbin:/bin:/usr/bin";
  const char *current_path = getenv ("PATH");

  char *path = NULL;

  LOG_MSG ("PATH=%s\n", current_path);

  /*
   * On glibc_small-based kernels the usermode-agent has no PATH -
   * until this is either fixed or understood we force a path which
   * will work.
   */
  path = format_msg ("%s:%s", default_path,
		     current_path ? current_path : "");

  assert (path);

  snprintf (cmd, sizeof cmd, "PATH=%s %s 2>&1",
	    path, modprobe_lttcontrol_cmd);

  if (path)
    free (path);

  if ((fp = popen (cmd, "r")) == NULL)
    die (EXIT_FAILURE, "%s: %s\n", cmd, strerror (errno));

  LOG_MSG ("executing: /bin/sh -c %s\n", cmd);

  while (fgets (buf, sizeof buf, fp))
    {
      /* only expecting errors so log to the error stream. */

      int n = strlen (buf);

      if (n > 0)
	buf[n-1] = '\0';

      LOG_ERR ("%s\n", buf);
    }

  rc = pclose (fp);

  return rc;
}


/*
 * umount the relayfs filesystem, assumed to be mounted at
 * RELAYFS_MOUNTPOINT. Returns OK on success, otherwise ERROR with errno
 * set appropriately.
 */
#if 0
static int
umount_relayfs (void)
{
  if (umount (relayfs_mountpoint) < 0)
    LOG_MSG ("umount: %s: %s\n", relayfs_mountpoint, strerror (errno));

  return OK;
}
#endif


/*
 * Mount the relayfs filesystem at MOUNTPOINT. Returns OK on success,
 * otherwise ERROR (with errno set appropriately).
 */
static int
mount_relayfs (const char *mountpoint)
{
  int rc = 0;
  int mountflags = 0;

  rc = mount ("nodev", mountpoint, "relayfs", mountflags, NULL);

  if (rc == 0 || (rc == -1 && errno == EBUSY))
    return OK;

  return ERROR;
}

/**
 * Returns TRUE if "ltt_control" is matched in /proc/modules, otherwise
 * FALSE.
 */
static int
lttcontrol_module_loaded_p (void)
{
  int retval = FALSE;
  char line[4096];			/* line of data  */
  FILE *fp = fopen ("/proc/modules", "r");

  if (!fp)
    {
      LOG_ERR ("cannot open /proc/modules: %s\n", strerror (errno));
      return FALSE;
    }

  while (fgets (line, sizeof line, fp))
    {
      if (strstr (line, "ltt_control") != NULL ||
	  strstr (line, "ltt-control") != NULL)
	{
	  retval = TRUE;
	  break;
	}
    }

  fclose (fp);

  return retval;
}


/**
 * Checks to see if the ltt-control module is loaded. Returns OK on
 * success, otherwise calls die() with an error message.
 */
static int
check_for_ltt_control_kernel_module (void)
{
  LOG_MSG ("checking for ltt-control module\n");

  if (!lttcontrol_module_loaded_p ())
    {
      if (modprobe_lttcontrol () != 0)
	  {
	  die (EXIT_FAILURE,
	       "ltt-control kernel module not loaded; cannot continue\n");
	  }
    }

  return OK;
}


/*
 * Return TRUE if pathname exists and is a FIFO, otherwise FALSE.
 */
static int
fifo_exists_p (const char *pathname)
{
  struct stat sb;

  if (stat (pathname, &sb) == 0)
    return S_ISFIFO (sb.st_mode);
  else
    return FALSE;
}


/*
 * Return TRUE if pathname exists and is a directory, otherwise FALSE.
 */
static int
directory_exists_p (const char *pathname)
{
  struct stat sb;

  if (stat (pathname, &sb) == 0)
    return S_ISDIR (sb.st_mode);
  else
    return FALSE;
}


/**
 * Returns TRUE if the trace channel exists, otherwise FALSE.
 */
static int
ltt_trace_channel_exists_p (void)
{
  struct stat sb;
  return (stat (ltt_trace_pathname, &sb)) == 0;
}


/*
 * Checks to see if there's lttdaemon already running.  Returns TRUE if
 * there is another lttdaemon running, otherwise FALSE.  If 'FD' is
 * non-null then the file descriptor is set in *FD.  If 'FD' is null,
 * then the pipe is closed.
 */
static int
lttdaemon_is_already_running_p (int *fd)
{
  int pipe_wr_fd;

  if (!fifo_exists_p (lttdaemon_named_pipe))
    {
      return FALSE;
    }

  /*
   * If the FIFO does exist and we can connect to it, and not block,
   * then that implies there is another lttdaemon running.
   */
  pipe_wr_fd = open (lttdaemon_named_pipe, O_WRONLY | O_NONBLOCK);

  if (fd)
    *fd = pipe_wr_fd;
  else if (!fd && pipe_wr_fd != -1)
    close (pipe_wr_fd);

  /*
   * If the open succeeded then this means there is another lttdaemon
   * listening on the other end of the pipe.
   */

  return (pipe_wr_fd != -1);
}


/**
 * Create a file, like touch(1).  Returns OK on success, otherwise ERROR
 * (with errno set appropriately).
 */
#if 0
static int
touch_file (const char *pathname)
{
  FILE *fp = fopen (pathname, "w");

  assert (pathname);

  if (!fp)
    return ERROR;

  if (fclose (fp) != 0)
    return ERROR;

  return OK;
}
#endif


/**
 * Write the current pid, running/paused-status, logging-mode to
 * LTTDAEMON_STATUSFILE. Returns OK on success, otherwise ERROR.  Errors
 * are logged using LOG_ERR().
 */
static int
write_statusfile (void)
{
  char *str = 0;
  size_t str_len = 0;
  const char *logging_mode = 0;
  const char *running_status = 0;

  if (cliopt_flight_recorder_mode)
    logging_mode = "flight";
  else
    logging_mode = "normal";

  if (logging_status == STARTED)
    running_status = "started";
  else if (logging_status == STOPPED)
    running_status = "stopped";
  else
    assert (0);			/* this function is being called but
				   lttdaemon is not running */

  /* Open file by truncating */

  FILE *fp = fopen (lttdaemon_statusfile, "w");

  if (!fp)
    {
      LOG_ERR ("cannot open %s: %s\n", lttdaemon_statusfile, strerror (errno));
      return ERROR;
    }

  str = format_msg ("%d %s %s\n", getpid (), running_status, logging_mode);
  str_len = strlen (str);

  LOG_MSG ("write status file: %s", str); /* deliberately miss out \n */

  if (fwrite (str, 1, str_len, fp) != str_len)
    {
      LOG_ERR ("failed to write %d bytes to %s: %s\n",
	       str_len,
	       lttdaemon_statusfile,
	       strerror (errno));

      (void) fclose (fp);

      if (str)
	free (str);

      return ERROR;
    }

  if (str)
    free (str);

  if (fclose (fp) != 0)
    {
      LOG_ERR ("cannot close pid file: %s: %s\n",
	       lttdaemon_statusfile,
	       strerror (errno));
      return ERROR;
    }

  return OK;
}


/**
 * Write the current version of lttdaemon to
 * LTTDAEMON_VERSIONFILE. Returns OK on success, otherwise ERROR.
 * Errors are logged using LOG_ERR().
 */
static int
write_versionfile (void)
{
  char *str = 0;
  size_t str_len = 0;

  /* Open file by truncating */

  FILE *fp = fopen (lttdaemon_versionfile, "w");

  if (!fp)
    {
      LOG_ERR ("cannot open %s: %s\n",
	       lttdaemon_versionfile,
	       strerror (errno));
      return ERROR;
    }

  str = format_msg ("%s\n", wrlinux_version);
  str_len = strlen (str);

  LOG_MSG ("write version file: %s", str); /* deliberately miss out \n */

  if (fwrite (str, 1, str_len, fp) != str_len)
    {
      LOG_ERR ("failed to write %d bytes to %s: %s\n",
	       str_len,
	       lttdaemon_versionfile,
	       strerror (errno));

      (void) fclose (fp);

      if (str)
	free (str);

      return ERROR;
    }

  if (str)
    free (str);

  if (fclose (fp) != 0)
    {
      LOG_ERR ("cannot close file: %s: %s\n",
	       lttdaemon_versionfile,
	       strerror (errno));
      return ERROR;
    }

  return OK;
}


/**
 * Writes PID to LTTDAEMON_PIDFILE.  Returns OK on success, otheewise
 * ERROR.  All errors are logged using LOG_ERR().
 */
static int
write_pidfile (int pid)
{
  char *pidstr = 0;
  size_t pidstr_len = 0;

  /* Open pid file by truncating */

  FILE *fp = fopen (lttdaemon_pidfile, "w");

  if (!fp)
    {
      LOG_ERR ("cannot open %s: %s\n", lttdaemon_pidfile, strerror (errno));
      return ERROR;
    }

  pidstr = format_msg ("%d\n", pid);
  pidstr_len = strlen (pidstr);

  if (fwrite (pidstr, 1, pidstr_len, fp) != pidstr_len)
    {
      LOG_ERR ("failed to write %d bytes to %s: %s\n",
	       pidstr_len,
	       lttdaemon_pidfile,
	       strerror (errno));

      (void) fclose (fp);

      if (pidstr)
	free (pidstr);

      return ERROR;
    }

  if (pidstr)
    free (pidstr);

  if (fclose (fp) != 0)
    {
      LOG_ERR ("cannot close pid file: %s: %s\n",
	       lttdaemon_pidfile,
	       strerror (errno));
      return ERROR;
    }

  return OK;
}


/**
 * SIGINT and SIGTERM handler for start_logging().  When
 * a signal is received lttctl_stop() and lttctl_destroy() are issued
 * which become exit criteria for the loop in
 * start_logging().  Further signal handling is then
 * ignored as we don't want either of the lttctl_() functions to be
 * called multiple times.
 */
static void
logging_mode_signal_handler (int signum)
{
  static int sigint_count = 0;

  switch (signum)
    {
    case SIGHUP:
      have_sighup = 1;
      break;
    case SIGUSR1:
      have_sigusr1 = 1;
      break;
    case SIGUSR2:
      have_sigusr2 = 1;
      break;
    case SIGINT:
      /* fall through */
    case SIGTERM:
      have_sigint = 1;
      have_sigterm = 1;

      sigint_count++;

      if ((sigint_count % 3) == 0)
	{
	  /* If we hit SIGINT N times then reset the signal handler to
	   * the default (which calls exit()).  This allows this process
	   * to exit should things go horribly wrong; in theory this
	   * shouldn't be necessary because either the sigint (or
	   * sigterm) should be picked up in the while(1) loop in
	   * start_logging(). If it doesn't, then this mechanism will
	   * allow repeated SIGINT's to quit this process.
	   */
	  signal (signum, SIG_DFL);
	}

      break;
    }
}


/*
 * Open lttdaemon's named pipe.  Returns 0 on success, otherwise die()'s
 * on any error.
 */
static int
open_lttdaemon_named_pipe (int *fd)
{
  int pipe_rw_fd;

  pipe_rw_fd = open (lttdaemon_named_pipe, O_RDWR | O_NONBLOCK);

  if (fd)
    *fd = pipe_rw_fd;

  return pipe_rw_fd == -1 ? ERROR : OK;
}


/*
 * Create named pipe.  Return OK on success, ERROR on error (in which
 * case, errno is set appropriately).
 */
static int
create_lttdaemon_named_pipe (void)
{
  if (!fifo_exists_p (lttdaemon_named_pipe))
    {
      if (mkfifo (lttdaemon_named_pipe,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP) < 0)
	{
	  return ERROR;
	}
    }

  return OK;
}

static int
copy_file (const char *src, const char *dst)
{
  int rd_fd;
  int wr_fd;
  int saved_errno;
  char buf[8192];

  assert (src != dst);

  rd_fd = open (src, O_RDONLY);

  if (rd_fd == -1)
    {
      LOG_ERR ("cannot open %s: %s\n", src, strerror (errno));
      return -1;
    }

  wr_fd = open (dst, O_WRONLY | O_CREAT | O_TRUNC, 0666);

  if (wr_fd == -1)
    {
      saved_errno = errno;
      LOG_ERR ("cannot open %s: %s\n", dst, strerror (errno));
      goto error;
    }

  while (1)
    {
      int nread = read (rd_fd, buf, 8192);

      if (nread == -1)
	{
	  saved_errno = errno;
	  goto error;
	}

      if (nread == 0)
	break;

      if (TEMP_FAILURE_RETRY (write (wr_fd, buf, nread)) != nread)
	{
	  saved_errno = errno;
	  goto error;
	}
    }

  close (rd_fd);

  if (close (wr_fd) != 0)
    {
      saved_errno = errno;
      goto error;
    }

  return OK;

 error:
  if (rd_fd != -1)
    close (rd_fd);

  if (wr_fd != -1)
    close (wr_fd);

  errno = saved_errno;

  return ERROR;
}

/**
 * ftw callback for copy_eventdefs.  Returns 0 on success, otherwise
 * non-zero.
 */
static int
copy_eventdefs_ftw (const char *file,
		    const struct stat *sb, int flag)

{
  sb = sb;			/* remove compiler unused warning */
  flag = flag;			/* remove compiler unused warning */

  /*
   * Unfortunately ftw(3) doesn't give us any client data in the
   * callback, therefore we have to use the global `ltt_facilities_dir'.
   */

  if (flag == FTW_D)
    {
      int rc;
      char *newpath = 0;

      newpath = format_msg ("%s/eventdefs", cliopt_output_dir);

      LOG_MSG ("make directory %s\n", newpath);

      rc = make_directory_path (newpath);

      free (newpath);

      return (rc == OK) ? 0 : -1;
    }
  else if (flag == FTW_F)
    {
      int rc;
      char *newpath = 0;
      const char *p = strstr (file, ltt_facilities_dir);

      if (!p)
	{
	  /* We are not expecting this condition, so die if encountered. */
	  die (EXIT_FAILURE, "cannot find substring '%s' in '%s'\n",
	       file,
	       ltt_facilities_dir);
	}

      /* Search for the `cliopt_output_dir' in the current filename.
	 If they match exactly then file and cliopt_output_dir are at
	 the same level in the files to be copied.  Otherwise skip over
	 the leading trace hierarchy and append the basename of `file'
	 to the output path. */

      p += strlen (ltt_facilities_dir);

      assert (*p == '/');

      while (*p == '/')
	p++;

      newpath = format_msg ("%s/eventdefs/%s", cliopt_output_dir, p);

      LOG_MSG ("copy: %s %s\n", p, newpath);

      if ((rc = copy_file (file, newpath)) != OK)
	{
	  LOG_ERR ("cannot copy %s to %s: %s\n",
		   file,
		   newpath,
		   strerror (errno));
	}

      free (newpath);

      return (rc == OK) ? 0 : -1;
    }

  return 0;
}

/**
 * Copy all the event definition files from the LTTV facilities
 * directory.  Returns OK on success, otherwise ERROR.  All errors are
 * logged using LOG_ERR().
 */
static int
copy_eventdefs (void)
{
  return ftw (ltt_facilities_dir, copy_eventdefs_ftw, 255) == 0 ? OK : ERROR;
}

static int
open_channel_trace_pair (const char *channel_pathname,
			 const char *trace_pathname,
			 channel_trace_pair_t *p_channel_trace_pair)
{
  int fd1 = -1;
  int fd2 = -1;
  char *mmap_addr1 = 0;
  struct stat sb;

  assert (channel_pathname);
  assert (trace_pathname);

  fd1 = open (channel_pathname, O_RDONLY | O_NONBLOCK);

  if (fd1 == -1)
    {
      LOG_MSG ("cannot open: %s: %s\n", channel_pathname, strerror (errno));
      goto error;
    }

  /* mmap the channel */

  mmap_addr1 = mmap (0,
		     cliopt_subbuf_size * cliopt_nsubufs,
		     PROT_READ,
		     MAP_SHARED,
		     fd1,
		     0);

  if (mmap_addr1 == 0)
    {
      LOG_ERR ("cannot mmap: %s: %s\n", channel_pathname, strerror (errno));
      goto error;
    }

  if (cliopt_append_mode)
    fd2 = open (trace_pathname,
		O_RDWR | O_CREAT | O_APPEND, S_IRWXU | S_IRWXG | S_IRWXO);
  else
    fd2 = open (trace_pathname,
		O_RDWR | O_CREAT | O_TRUNC, S_IRWXU | S_IRWXG | S_IRWXO);

  if (fd2 == -1)
    {
      LOG_MSG ("cannot open %s: %s\n", trace_pathname, strerror (errno));
      goto error;
    }

  /*
   * If we are in append mode then we need to know the existing size of
   * * the file.  This size is subsequently used when writing to
   * trace_fd.
   */
  if (stat (trace_pathname, &sb) != 0)
    {
    LOG_ERR ("cannot stat: %s: %s\n", trace_pathname, strerror (errno));
    goto error;
    }

  p_channel_trace_pair->channel_fd = fd1;
  p_channel_trace_pair->trace_fd = fd2;
  p_channel_trace_pair->channel_mmap_addr = mmap_addr1;
  p_channel_trace_pair->hungup = 0;
  p_channel_trace_pair->channel_name = lttdaemon_strdup (channel_pathname);
  p_channel_trace_pair->trace_name = lttdaemon_strdup (trace_pathname);
  p_channel_trace_pair->trace_name_file_size = sb.st_size;
  p_channel_trace_pair->trace_fd_write_error = 0;

  return 0;

 error:

  if (fd1 != -1)
    (void) close (fd1);

  if (fd2 != -1)
    (void) close (fd2);

  if (mmap_addr1 != 0)
    (void) munmap (mmap_addr1, cliopt_subbuf_size * cliopt_nsubufs);

  return -1;
}


/**
 * Open all trace file and relayfs channels.  Returns 0 on success,
 * non-zero on error with all errors logged via LOG_ERR().
 */
static int
open_trace_files_ftw (const char *file,
		      const struct stat *sb, int flag)
{
  const char *q = 0;
  char *newpath = 0;

  sb = sb;			/* remove compiler unused warning */
  flag = flag;			/* remove compiler unused warning */

  /*
   * Unfortunately ftw(3) doesn't give us any client data in the
   * callback, therefore we have to use the global `ltt_trace_pathname'.
   */

  const char *p = strstr (file, ltt_trace_pathname);

  if (!p)
    {
      /* We are not expecting this condition, so die if encountered. */

      die (EXIT_FAILURE, "cannot find substring '%s' in '%s'\n",
	   ltt_trace_pathname, file);
    }

  /*
   * Search for the `ltt_trace_pathname' in the current filename.  If
   * they match exactly then file and ltt_trace_pathname are at the same
   * level in the files to be copied.  Otherwise skip over the leading
   * trace hierarchy and append the basename of `file' to the output
   * path.
   */
  if (strcmp (file, ltt_trace_pathname) == 0)
    {
      newpath = lttdaemon_strdup (cliopt_output_dir);
    }
  else
    {
      q = p + strlen (ltt_trace_pathname);

      while (*q == '/')
	q++;

      newpath = format_msg ("%s/%s", cliopt_output_dir, q);
    }

  if (flag == FTW_D)
    {
      LOG_MSG ("make directory %s\n", newpath);

      if (make_directory_path (newpath) != 0)
	return -1;
    }
  else if (flag == FTW_F)
    {
      channel_trace_pair_t *p_channel_trace_pair =
	&channel_trace_pair[channel_trace_pair_count];

      if (open_channel_trace_pair (file,
				   newpath, p_channel_trace_pair) != 0)
	return -1;

      LOG_MSG ("open %s\n", newpath);

      channel_trace_pair_count++;
    }

  if (newpath)
    free (newpath);

  return 0;
}


/**
 * Opens and mmaps all files in LTT_TRACE_PATHNAME.  Returns OK on
 * success, otherwise ERROR if any file could not be opened or mmap'ed.
 * Errors are logged using LOG_ERR().
 */
static int
open_trace_files (void)
{
  return ftw (ltt_trace_pathname, open_trace_files_ftw, 255);
}

/**
 * Close the filehandles and mmap arenas held open by the channel.
 *
 * BUG: ALWAYS RETURNS OK.
 */
static int
close_channel_trace_pair (channel_trace_pair_t *p)
{
  (void) close (p->channel_fd);
  (void) close (p->trace_fd);

  if (p->channel_mmap_addr)
    (void) munmap (p->channel_mmap_addr, cliopt_subbuf_size);

  p->channel_fd = -1;
  p->trace_fd = -1;
  p->channel_mmap_addr = NULL;

  return OK;
}

/**
 * Close all trace channels.
 */
#if 0
static int
close_all_channel_trace_pairs (void)
{
  int i;

  for (i = 0; i < channel_trace_pair_count; i++)
    {
      if (close_channel_trace_pair (&channel_trace_pair[i]) != OK)
	return ERROR;
    }

  return OK;
}
#endif


/**
 * Consume data from trace channel.  Returns OK on success, otherwise
 * ERROR.
 */
static int
ltt_read_subbuffer (channel_trace_pair_t *pair)

{
  static char *bigbuf = NULL;

  unsigned int old = 0;
  size_t n;
  int ioctl_err = 0;

  if (bigbuf == NULL)
    bigbuf = lttdaemon_malloc (cliopt_subbuf_size);

  if (ioctl (pair->channel_fd, RELAYFS_GET_SUBBUF, &old) != 0)
    {
      LOG_ERR ("cannot reserve sub buffer: %s: %s\n",
	       channel_trace_pair->channel_name,
	       strerror (errno));

      return ERROR;
    }

/*   printf ("%d: cookie : %u\n", pair->channel_fd, old); */

  if (!pair->trace_fd_write_error)
    {
      /*
       * Although this would appear to the /most/ expensive way of doing
       * this (i.e., why not just do the write straightaway), it reduces
       * the number of times the error message "reader has been pushed
       * by the writer" occurs.  So it may actually be more efficient.
       * I don't fully understand the ramifications of "reader pushed by
       * the writer".  Should we stop logging?  Should we return an
       * error?  As things stand today, we simply don't write the data
       * when this happens.
       */

      memcpy (bigbuf, ((char *) pair->channel_mmap_addr +
		       (old & ((cliopt_nsubufs * cliopt_subbuf_size) -1))),
	      cliopt_subbuf_size);
    }

  if ((ioctl_err = ioctl (pair->channel_fd, RELAYFS_PUT_SUBBUF, &old)) != 0)
    {
      switch (errno)
	{
	case EFAULT:
	  LOG_ERR ("cannot unreserve sub buffer: %s: %s\n",
		   pair->channel_name,
		   strerror (errno));
	  return ERROR;
	case EIO:
	  LOG_WARN ("\
reader has been pushed by the writer, last subbuffer corrupted; \
data not copied: %s: %s\n",
		    pair->channel_name,
		    strerror (errno));

	  /* FIXME : what should happen here?  We haven't written bad
	     data as such. */
	  return OK;
	default:
	  LOG_ERR ("ltt_read_subbuffer: %s\n", strerror (errno));
	  return ERROR;
	}
    }

  if (pair->trace_fd_write_error)
    return OK;

  n = TEMP_FAILURE_RETRY(write (pair->trace_fd, bigbuf, cliopt_subbuf_size));

  if (n != cliopt_subbuf_size)
    {
      LOG_ERR ("cannot write: %s: %s\n", pair->trace_name, strerror (errno));
      return ERROR;
    }

  return OK;
}


/**
 * Install various signal handlers when in logging mode.  Returns OK on
 * success, ERROR.  All errors are logged via LOG_ERR().
 */
static int
install_logging_mode_signal_handlers (void)
{
  /*
   * Use the signal() call to associate our own functions with the
   * SIGINT and SIGTERM signals.  For this collection mode we want to
   * bail out of the while(1) loop.
   */

  if (signal (SIGHUP, logging_mode_signal_handler) == SIG_ERR)
    {
      LOG_ERR ("cannot install SIGHUP handler: %s\n", strerror (errno));
      return ERROR;
    }

  if (signal (SIGINT, logging_mode_signal_handler) == SIG_ERR)
    {
      LOG_ERR ("cannot install SIGINT handler: %s\n", strerror (errno));
      return ERROR;
    }

  if (signal (SIGTERM, logging_mode_signal_handler) == SIG_ERR)
    {
      LOG_ERR ("cannot install SIGTERM handler: %s\n", strerror (errno));
      return ERROR;
    }

  if (signal (SIGUSR1, logging_mode_signal_handler) == SIG_ERR)
    {
      LOG_ERR ("cannot install SIGUSR1 handler: %s\n", strerror (errno));
      return ERROR;
    }

  if (signal (SIGUSR2, logging_mode_signal_handler) == SIG_ERR)
    {
      LOG_ERR ("cannot install SIGUSR2 handler: %s\n", strerror (errno));
      return ERROR;
    }

  return OK;
}

/**
 * Stop LTT logging and close all the trace channels.
 */
static void
stop_logging (void)
{
  static int processed = 0;

  if (!processed)		/* only stop and destroy once */
    {
      LOG_MSG ("stop logging\n");

      (void) lttctl_stop (lttctl_handle, cliopt_trace_name);

      /*
       * By destroying the channels we get a POLLHUP for each trace FD
       * we are listening on.  When all trace FDs receive a POLLHUP
       * that's our exit criteria in start_logging().
       */

      (void) lttctl_destroy_trace (lttctl_handle, cliopt_trace_name);
      processed = 1;
    }
}

/**
 * Check to see if a signal was received.  Returns OK on success,
 * otherwise ERROR.
 */
static int
check_for_signal_receipt (void)
{
  /* sigusr1 == stop logging */

  if (have_sigusr1)
    {
      have_sigusr1 = 0;	/* clear flag */

      if (logging_status == STARTED)
	{
	  if (lttctl_stop (lttctl_handle, cliopt_trace_name) != 0)
	    {
	      LOG_ERR ("unable to pause logging: %s\n", strerror (errno));
	      return ERROR;
	    }

	  if (write_statusfile () != 0)
	    return ERROR;
	}
      else
	{
	  LOG_MSG ("ignoring request; already stopped\n");
	}
    }

  /* sigusr2 == resume logging */

  if (have_sigusr2)
    {
      have_sigusr2 = 0;	/* clear flag */

      if (logging_status == STOPPED)
	{
	  if (lttctl_start (lttctl_handle, cliopt_trace_name) != 0)
	    {
	      LOG_ERR ("unable to resume logging: %s\n", strerror (errno));
	      return ERROR;
	    }

	  if (write_statusfile () != 0)
	    return ERROR;
	}
      else
	{
	  LOG_MSG ("ignoring request; already logging\n");
	}
    }

  /* sigint || sigterm == stop logging and collect log requested */

  if (have_sigint || have_sigterm)
    {
      have_sigint  = 0;	/* clear flag */
      have_sigterm = 0;	/* clear flag */

      (void) stop_logging ();
    }

  /* sighup == write status file. */

  if (have_sighup)
      {
      have_sighup = 0;
      (void) write_statusfile ();
      }

  return OK;
}


/**
 * Start logging. This is similar to lttd(1).  Runs until a "collection"
 * request is issued.  Returns OK on success, otherwise ERROR.  All
 * errors are logged via LOG_ERR().
 */
static int
start_logging (void)
{
  int i;
  int rc;
  int num_hup = 0;
  int trace_files_opened = 0;
  int poll_timeout = -1;
  struct pollfd pollfds[MAX_TRACE_FILES];

  if (install_logging_mode_signal_handlers () != OK)
    return ERROR;

  if (cliopt_flight_recorder_mode)
    {
    LOG_MSG ("start logging in flight recorder mode\n");
    }
  else
    {
    LOG_MSG ("start logging\n");
    }

  (void) MapSlashProc ();

  if (lttctl_start (lttctl_handle, cliopt_trace_name) != 0)
    return ERROR;

  if (write_pidfile (getpid ()) != 0)
    die (EXIT_FAILURE, NULL);

  if (write_statusfile () != 0)
    die (EXIT_FAILURE, NULL);

  while (1)
    {
      /*
       * Only open the files in normal mode logging, and only open them
       * once.  When we switch from flight recorder mode to normal mode
       * the files will be opened.  And when running in flight recorder
       * mode we don't want to open them until collection has been
       * signalled.
       */
      if (cliopt_normal_mode && !trace_files_opened)
	{
	  /*
	   * Copy and mimic the trace channel hierarchy to the
	   * cliopt_output_dir.  This copies:
	   * /mnt/relayfs/ltt/trace/cpu_0 to /output-dir/cpu_0
	   * /mnt/relayfs/ltt/trace/control to /output-dir/control
	   * /mnt/relayfs/ltt/trace/control/facilities_0 to /output-dir/control/facilities_0
	   * etc.
	   */
	  if (open_trace_files () != 0)
	    return ERROR;

	  /*
	   * Populate the FDs to listen on based on those that have not hungup.
	   */
	  for (i = 0; i < channel_trace_pair_count; i++)
	    {
	      if (i == MAX_TRACE_FILES)
		{
		die (EXIT_FAILURE, "only %d trace files are supported\n",
		     MAX_TRACE_FILES);
		}

	      assert (i < MAX_TRACE_FILES);

	      pollfds[i].fd = channel_trace_pair[i].channel_fd;
	      pollfds[i].events = POLLIN | POLLPRI;
	    }

	  trace_files_opened = 1;
	}

      /*
       * Handle all signals up front.  This allows pause/resume to work
       * in flight recorder mode.
       */

      if (check_for_signal_receipt () != 0)
	return ERROR;

      if (cliopt_flight_recorder_mode)
	{
	  rc = poll (NULL, 0, -1);

	  /*
	   * Wait until we're signalled by the signal handler; once
	   * signalled we'll carry on logging in the normal way.
	   */
	  if (rc == -1 && errno == EINTR && (have_sigint || have_sigterm))
	    {
	      /* Switch to normal mode logging. */

	      LOG_MSG ("switching to normal mode logging\n");

	      cliopt_normal_mode = 1;
	      cliopt_flight_recorder_mode = 0;

	      continue;		/* call open trace files. */
	    }
	  else if (rc == -1 && errno != EINTR)
	    {
	      LOG_ERR ("poll: %s\n", strerror (errno));
	      return ERROR;
	    }
	}
      else
	{
	  rc = poll (pollfds, channel_trace_pair_count, poll_timeout);

	  if (rc == -1)
	    {
	      if (errno == EINTR)
		{
		  continue;
		}
	      else
		{
		  LOG_ERR ("poll: %s\n", strerror (errno));
		  return ERROR;
		}
	    }
	  else if (rc == 0)
	    {
	      LOG_MSG ("poll timeout!\n");
	      continue;			     /* timeout */
	    }

	  assert (rc > 0);

	  for (i = 0; i < rc; i++)
	    {
	      switch (pollfds[i].revents)
		{
		case POLLERR:
		  if (channel_trace_pair[i].hungup == 0)
		    {
		      LOG_MSG ("Error returned in polling fd %d\n",
			       pollfds[i].fd);
		      channel_trace_pair[i].hungup = 1;
		      num_hup++;
		    }
		  break;
		case POLLHUP:
		  if (channel_trace_pair[i].hungup == 0)
		    {
		      LOG_MSG ("polling fd %d tells it has hung up\n",
			       pollfds[i].fd);
		      channel_trace_pair[i].hungup = 1;
		      num_hup++;
		    }
		  break;
		case POLLNVAL:
		  if (channel_trace_pair[i].hungup == 0)
		    {
		      LOG_MSG ("Polling fd %d tells fd is not open\n",
			       pollfds[i].fd);
		      channel_trace_pair[i].hungup = 1;
		      num_hup++;
		    }
		  break;
		case POLLIN:
		case POLLPRI:
		  if (ltt_read_subbuffer (&channel_trace_pair[i]) != OK)
		    {
		      channel_trace_pair[i].trace_fd_write_error = 1;

		      /*
		       * By stopping logging and closing the channel
		       * we're bound to (!!) catch one of POLLHUP |
		       * POLLNVAL - this helps in our quest to exit this
		       * while(1) loop.
		       */
		      (void) stop_logging ();
		      (void) close_channel_trace_pair (&channel_trace_pair[i]);
		    }
		  break;
		}
	    }

	  if (num_hup == channel_trace_pair_count)
	    {
	      LOG_MSG ("every FD has hung up\n");
	      break;
	    }
	}
    }

  LOG_MSG ("collection done!\n");

  return (errors == 0) ? OK : ERROR;
}


/**
 * Opens the ltt control channel. Returns non-NULL on success.
 */
static struct lttctl_handle *
open_lttctrl_channel (void)
{
  return lttctl_create_handle();
}


/**
 * Returns the pid for an existing lttdaemon process.  Errors are raised
 * through die() so the only return path is for the file to exist, be a
 * valid integer and `kill(pid, 0)' returns OK.
 */
static int
get_running_lttdaemon_pid (void)
{
  int pid = -1;
  char buf[512];
  int len = 0;
  FILE *fp = fopen (lttdaemon_pidfile, "r");

  if (!fp)
    {
      die (EXIT_FAILURE, "cannot open: %s: %s\n",
	   lttdaemon_pidfile, strerror (errno));
    }

  buf[0] = '\0';

  if (fgets (buf, 512, fp) == NULL)
    {
      die (EXIT_FAILURE, "cannot read pid value from %s: %s\n",
	   lttdaemon_pidfile,
	   strerror (errno));
    }

  (void) fclose (fp);

  len = strlen (buf);

  if (len > 0 && buf[len -1] == '\n')
    buf[len-1] = '\0';

  /* convert string representation to int. */

  if (lttdaemon_atoi (buf, &pid) != OK)
    {
      die (EXIT_FAILURE, "pid value '%s' from %s is not a valid integer\n",
	   buf,
	   lttdaemon_pidfile);
    }

  if (kill (pid, 0) < 0)
    {
      die (EXIT_FAILURE, "lttdaemon (pid %d) is not running: %s\n",
	   pid, strerror (errno));
    }

  return pid;
}

static int
create_tracefile_list_ftw (const char *file,
			   const struct stat *sb, int flag)
{
  char *line = NULL;
  const char *p = strstr (file, cliopt_output_dir);

  if (!p)
    {
      /* We are not expecting this condition, so die if encountered. */
      die (EXIT_FAILURE, "cannot find substring '%s' in '%s'\n",
	   cliopt_output_dir,
	   file);
    }

  if (flag == FTW_F)
    {
      assert (p);

      p += strlen (cliopt_output_dir);

      while (*p == '/')
	p++;

      line = format_msg ("F\t%ld\t%s\n", sb->st_size, p);
    }
  else if (flag == FTW_D)
    {
      if (strcmp (file, cliopt_output_dir) == 0)
	return 0;		/* skip cliopt_output_dir */

      p += strlen (cliopt_output_dir);

      while (*p == '/')
	p++;

      line = format_msg ("D\t%s\n", p);
    }

  if (flag == FTW_F || flag == FTW_D)
    {
      int rc;
      int saved_errno;
      size_t n = 0;

      assert (line);
      n = strlen (line);
      rc = fwrite (line, 1, n, file_list_stream);
      saved_errno = errno;
      free (line);
      errno = saved_errno;

      return (rc == (int) n) ? 0 : -1;
    }

  return 0;
}

static int
create_tracefile_list (void)
{
  int rc;
  char *s = format_msg ("%s/wrsv.filelist", cliopt_output_dir);

  assert (s);

  if ((file_list_stream = fopen (s, "w")) == NULL)
    {
      free (s);
      LOG_ERR ("cannot open: %s: %s\n", s, strerror (errno));
      return ERROR;
    }

  if ((rc = ftw (cliopt_output_dir, create_tracefile_list_ftw, 255)) != 0)
    {
      LOG_ERR ("write failed: %s: %s\n", s, strerror (errno));
      free (s);
      return ERROR;
    }

  if (fclose (file_list_stream) != 0)
    {
      LOG_ERR ("cannot close: %s: %s\n", s, strerror (errno));
      free (s);
      return ERROR;
    }

  if (s)
    free (s);

  return (rc == 0) ? OK : ERROR;
}


/**
 * A rather naive pathname canonicalizer.  It removes both tailing /
 * characters as well as repeated / characters in the middle of
 * PATHNAME.  Returns a new string if successful, otherwise NULL.  I'm
 * sure there's a cheaper, faster, better way of doing this.
 */
#if 0
static char *
canonicalized_pathname (const char *pathname)
{
  char *src_copy = lttdaemon_strdup (pathname);
  char *dst = lttdaemon_malloc (strlen (pathname) + 1);
  char *dst_orig = dst;
  char *result = NULL;
  int i = 0;
  int j = 0;
  int src_len = 0;

  /* coalesce trailing / characters. */

  while (1)
    {
      int x = strlen (src_copy);

      if (x > 0 && src_copy[x-1] == '/')
	src_copy[x-1] = '\0';
      else
	break;
    }

  src_len = strlen (src_copy) + 1; /* copy null as well */

  /* trim repeated / characters. */

  for (i = 0, j = 0; i < src_len; i++)
    {
      if (!(j > 0 && src_copy[i] == '/' && dst[j-1] == '/'))
	dst[j++] = src_copy[i];
    }

  /* return a copy of the shorter string. */

  result = lttdaemon_strdup (dst_orig);

  /* free working copies. */

  free (src_copy);
  free (dst);

  return result;
}
#endif

static void
print_usage_and_die (void)
{
  LOG_ERR ("\
usage: %s [ARGS]\n\
ARGS:\n\
 -a --- append mode\n\
 -c --- collect log\n\
 -f --- flight recorder mode\n\
 -n --- number of subbuffers (min 2, max %d)\n\
 -o --- output directory\n\
 -p --- pause logging\n\
 -q --- query daemon status\n\
 -r --- resume logging\n\
 -v --- verbose mode\n\
 -x --- size of subbuffer (min %dKB, max %dMB)\n\
\n\
Examples:\n\
\n\
 Start logging, output is in directory /tmp/foo\n\
     lttdaemon -o /tmp/foo [-a] [-n <n>] [-x <x>]\n\
\n\
 Start logging in flight recorder mode, output is in directory /tmp/foo\n\
     lttdaemon -o /tmp/foo [-a] [-n <n>] [-x <x>] -f\n\
\n\
 Stop logging and collect log\n\
     lttdaemon -c\n\
\n\
 Pause logging\n\
     lttdaemon -p\n\
\n\
 Resume logging\n\
     lttdaemon -r\n\
\n\
 Query status\n\
     lttdaemon -q\n\
\n\
A SIGTERM or SIGINT will also stop logging.  It is the equivalent\n\
of `lttdaemon -c', but only when logging is in effect.\n\
\n\
For example, running lttdaemon from the shell:\n\
\n  # lttdaemon -o /tmp/foo  [Ctrl-C] in the shell will stop logging.\n\
\n\
Invocations of %s communicate through %s\n\
The executing status is written to %s on either SIGHUP or -q\n\
\nversion: %s\n",
	   progname,
	   max_subbufs,
	   getpagesize(),
	   max_subbuf_size / (1024 * 1024),
	   progname,
	   lttdaemon_named_pipe,
	   lttdaemon_statusfile,
	   wrlinux_version);

  lttdaemon_exit (EXIT_FAILURE);
}

/**
 * Returns next power-of-2 from N.
 */
static unsigned int
round_up_to_next_power_of_2 (unsigned int n)
{
    unsigned int i;

    n &= INT_MAX;

    for (i = 1; n > i; i <<= 1)
	{}

    return i;
}


/**
 * Process and verify command line options.  All options are reflected
 * in cliopt_* global parameters.  Returns OK on success, ERROR on error
 * with '*NONOPT_INDEX' set to the first non option index in argv.  Any
 * errors are logged via LOG_ERR().
 */
static int
process_command_line_options (int argc, char *argv[], int *nonopt_index)
{
  int c;
  int have_aflag = 0;
  int have_cflag = 0;
  int have_fflag = 0;
  int have_nflag = 0;
  int have_oflag = 0;
  int have_pflag = 0;
  int have_rflag = 0;
  int have_tflag = 0;
  int have_xflag = 0;
  int have_vflag = 0;
  int have_qflag = 0;
  int have_mflag = 0;

  /*
   * Prevent getopt from printing error message; we want to handle them
   * ourselves.
   */
  opterr = 0;

  while ((c = getopt (argc, argv, "acfhmn:o:pqrt:x:v?")) != -1)
    {
      switch (c)
	{
	case 'a':		/* collect log in append mode */
	  have_aflag = 1;
	  cliopt_append_mode = 1;
	  break;

	case 'c':		/* collect log */
	  have_cflag = 1;
	  break;

	case 'f':		/* flight recorder mode */
	  have_fflag = 1;
	  break;

	case 'm':		/* ltt-control is a kernel module */
	  have_mflag = 1;
	  cliopt_load_ltt_control_module = 1;
	  break;

	case 'n':		/* number of subbufs */
	  {
	    int original_val = 0;

	    have_nflag = 1;

	    if (lttdaemon_atoi (optarg, &cliopt_nsubufs) != OK)
	      {
		LOG_ERR ("\
-n: invalid argument: `%s': must be a number\n", optarg);
		return ERROR;
	      }

	    if (errno == ERANGE)
	      {
		LOG_ERR ("\
-n: invalid argument: `%s': must be a number > 0 and <= %d\n",
			 optarg,
			 max_subbufs);
		return ERROR;
	      }

	    original_val = cliopt_nsubufs;

	    if (cliopt_nsubufs < 2)
	      cliopt_nsubufs = 2;

	    cliopt_nsubufs = round_up_to_next_power_of_2 (cliopt_nsubufs);

	    if (cliopt_nsubufs >= max_subbufs)
		cliopt_nsubufs = max_subbufs;

	    if (original_val != cliopt_nsubufs)
	      {
		LOG_WARN("-n %d: rounded/clamped to %d\n",
			 original_val,
			 cliopt_nsubufs);
	      }

	    break;
	  }
	case 'o':
	  have_oflag = 1;
	  cliopt_output_dir = optarg;
/*	  cliopt_output_dir = canonicalized_pathname (optarg); */
	  break;

	case 'p':
	  have_pflag = 1;
	  break;

	case 'q':
	  have_qflag = 1;
	  cliopt_query_status = 1;
	  break;

	case 'r':
	  have_rflag = 1;
	  break;

	case 't':
	  have_tflag = 1;
	  cliopt_trace_name = optarg;
	  break;

	case 'x':		/* size of subbufs */
	  {
	    int original_val = 0;
	    int pagesize = getpagesize ();
	    int xval;

	    have_xflag = 1;

	    if (lttdaemon_atoi (optarg, &xval) != OK)
	      {
		LOG_ERR ("\
-x: invalid argument: `%s': must be a number\n",
			 optarg);
		return ERROR;
	      }

	    if (errno == ERANGE)
	      {
		LOG_ERR ("\
-x: invalid argument: `%s': must be a number >= %d and <= %d\n",
			 optarg,
			 getpagesize (),
			 max_subbuf_size);
		return ERROR;
	      }

	    original_val = xval;

	    /* Validate bounds of original request. */

	    if ((int) xval < pagesize)
	      xval = pagesize;

	    if ((int) xval > max_subbuf_size)
	      xval = max_subbuf_size;

	    /* Round up to next power-of-2 (pagesize) boundary. */

	    xval = round_up_to_next_power_of_2 (xval);

	    assert (xval >= pagesize);
	    assert ((xval % pagesize) == 0);

	    if (original_val != xval)
	      {
		LOG_WARN("-x %d: rounded/clamped to %u\n",
			 original_val,
			 xval);
	      }

	    cliopt_subbuf_size = xval;

	    break;
	  }
	case 'v':
	  have_vflag = 1;
	  cliopt_verbose = 1;
	  break;

	case '?':
	case 'h':
	  print_usage_and_die ();
	  return ERROR;

	default:
	  if (isprint (optopt))
	    {
	      LOG_ERR ("unknown option `-%c'.\n", optopt);
	      print_usage_and_die ();
	    }
	  else
	    {
	      LOG_ERR ("unknown option character `\\x%x'.\n", optopt);
	      print_usage_and_die ();
	    }

	  return ERROR;
	}
    }

  *nonopt_index = optind;	/* parameter result */

  /* Validate pause and resume combinations. */

  if (have_pflag || have_rflag)
    {
      if (have_pflag && have_rflag)
	{
	  LOG_ERR ("-p and -r are mutually exclusive\n");
	  return ERROR;
	}

      if (have_pflag && (have_aflag ||
			 have_cflag ||
			 have_fflag ||
			 have_nflag ||
			 have_oflag ||
			 have_qflag ||
			 have_rflag ||
			 have_tflag ||
			 have_xflag))
	{
	  LOG_ERR ("cannot use -p mode with other options\n");
	  return ERROR;
	}
      else if (have_rflag && (have_aflag ||
			      have_cflag ||
			      have_fflag ||
			      have_nflag ||
			      have_oflag ||
			      have_pflag ||
			      have_qflag ||
			      have_tflag ||
			      have_xflag))
	{
	  LOG_ERR ("cannot use -r with other options\n");
	  return ERROR;
	}
      else
	{
	  if (have_pflag)
	    cliopt_pause = 1;
	  else if (have_rflag)
	    cliopt_resume = 1;
	  else
	    assert (0);

	  return OK;
	}
    }

  if (have_cflag)
    {
      if (have_aflag ||
	  have_fflag ||
	  have_mflag ||
	  have_nflag ||
	  have_oflag ||
	  have_pflag ||
	  have_qflag ||
	  have_rflag ||
	  have_tflag ||
	  have_xflag)
	{
	  LOG_ERR ("\
cannot specify other options when using -c\n");
	  return ERROR;
	}
      else
	{
	  cliopt_collect_log = 1;
	  return OK;
	}
    }

  if (have_oflag)
    {
      if ((have_pflag || have_qflag || have_rflag || have_tflag))
	{
	  LOG_ERR ("valid options for -o are: -a, -n, -x, -f\n");
	  return ERROR;
	}
      else
	{
	  if (have_fflag)
	    cliopt_flight_recorder_mode = 1;
	  else
	    cliopt_normal_mode = 1;

	  return OK;
	}
    }

  if (have_qflag)
    return OK;

  print_usage_and_die ();

  return ERROR;
}


int
main (int argc, char *argv[])
{
  int rc;
  int nonopt_index;
  char resolved_path[PATH_MAX];

  /*
   * We want all errors and messages logged ASAP as we are not running
   * in the context of a shell.  This process is being invoked via the
   * usermode-agent and our only hope of detecting anomalies/errors is
   * to record them somewhere so that they may subsequently be retrieved
   * (by the UI).  Opening the logging streams must therefore always
   * come first.
   */
  if ((rc = open_logfile (progname)) != 0)
    lttdaemon_exit (rc);

  /*
   * The atexit handler deletes the logging files if there were no
   * errors recorded.
   */
  if (atexit (atexit_handler))
    die (EXIT_FAILURE, "cannot set exit handler: %s\n", strerror (errno));

  /* We use mount(1) and modprobe(1)... so we have to be root. */

  if (getuid () != 0)
    die (EXIT_FAILURE, "must be root!\n");

  if (!directory_exists_p (ltt_facilities_dir))
    {
      LOG_ERR ("cannot find facilities directory: %s: %s\n",
	       ltt_facilities_dir,
	       strerror (errno));

      die (EXIT_FAILURE, "the LTT package needs to be installed\n");
    }

  /* Process command line options. */

  if (process_command_line_options (argc, argv, &nonopt_index) != OK)
    die (EXIT_FAILURE, "failed to parse options\n");

  if (write_versionfile () != OK)
    {
      die (EXIT_FAILURE, "cannot write version file: %s\n",
	   lttdaemon_versionfile);
    }

  /* Process requests. */

  if (cliopt_flight_recorder_mode || cliopt_normal_mode)
    {
      int pipe_rw_fd;

      if (lttdaemon_is_already_running_p (NULL))
	die (EXIT_FAILURE, "lttdaemon is already running\n");

      if (make_directory_path (relayfs_mountpoint) != OK)
	{
	  die (EXIT_FAILURE, NULL);
	}

      if (mount_relayfs (relayfs_mountpoint) != OK)
	{
	  die (EXIT_FAILURE, "cannot mount: %s: %s\n",
	       relayfs_mountpoint, strerror (errno));
	}

      /*
       * Must have ltt-control module loaded for open_lttctrl_channel()
       * to work.  XXX always a module for the moment!
       */
      if (1 || cliopt_load_ltt_control_module)
	check_for_ltt_control_kernel_module ();

      /*
       * Pretty much every operation requires an ltt-control handle so do it
       * upfront.
       */
      if ((lttctl_handle = open_lttctrl_channel ()) == NULL)
	die (EXIT_FAILURE, "cannot create ltt control handle!\n");

      /*
       * If there is no existing daemon running then clean up any
       * previous run by deleting the trace channel.  If the trace
       * channel exists then creating/starting tracing will
       * subsequently fail with 'channel already exists'.
       */
      if (ltt_trace_channel_exists_p ())
	{
	  if (lttdaemon_stop_destroy_trace () != 0)
	    {
	      die (EXIT_FAILURE, "cannot cleanup existing trace channel: %s\n",
		   ltt_trace_pathname);
	    }
	}

      /*
       * Create/open the named pipe.  This is how subsequent invocations
       * of this program detect that there is a real/live lttdaemon
       * still running.
       */
      if (create_lttdaemon_named_pipe () != OK)
	{
	  die (EXIT_FAILURE, "cannot create FIFO %s: %s\n",
	       lttdaemon_named_pipe,
	       strerror (errno));
	}

      if (open_lttdaemon_named_pipe (&pipe_rw_fd) != OK)
	{
	  die (EXIT_FAILURE, "cannot open named pipe: %s: %s\n",
	       lttdaemon_named_pipe, strerror (errno));
	}

      /* Create the relayfs trace channels. */

      if ((rc = lttctl_create_trace (lttctl_handle,
				     cliopt_trace_name,
				     LTT_TRACE_NORMAL,
				     cliopt_subbuf_size,
				     cliopt_nsubufs)) != 0)
	die (EXIT_FAILURE, "\
create trace failed: \
the number of buffers and their size may be too large for your kernel\n");

      LOG_MSG ("copying eventdefs\n");

      /*
       * First make the directory so that we can canonicalize
       * cliopt_output_dir.  It needs to be in a canonical form as
       * copy_eventdefs() relies on this.
       */

      if (make_directory_path (cliopt_output_dir) != 0)
	{
	  die (EXIT_FAILURE, NULL);
	}

      /* Canonicalize path. */
      if (realpath ((const char*) cliopt_output_dir, resolved_path) == NULL)
	{
	  die (EXIT_FAILURE, "realpath: %s: %s\n",
	       cliopt_output_dir, strerror (errno));
	}

      cliopt_output_dir = lttdaemon_strdup (resolved_path);

      /*
       * Copy the eventdefs *.xml files to the output directory.
       */
      if (copy_eventdefs () != 0)
	die (EXIT_FAILURE, NULL);

      rc = start_logging ();

      (void) unlink (lttdaemon_pidfile);
      (void) unlink (lttdaemon_statusfile);
      (void) unlink (lttdaemon_named_pipe);
      (void) lttctl_destroy_handle (lttctl_handle);

      if (rc == 0)
	{
	  if (create_tracefile_list () != OK)
	    die (EXIT_FAILURE, NULL);
	}

      if (errors > 0)
	lttdaemon_exit (EXIT_FAILURE);

      if (warnings > 0)
	lttdaemon_exit (EXIT_STATUS_WITH_WARNINGS);

      if (cliopt_output_dir)
	{
	  free (cliopt_output_dir);
	  cliopt_output_dir = NULL;
	}

      lttdaemon_exit (rc);
    }
  else
    {
      /*
       * This is not equivalent to getpid(). This is the pid of the already
       * running daemon process.
       */
      int lttdaemon_pid = -1;

      if (cliopt_query_status)
	{
	  /* Don't log an error for the query command. */

	  lttdaemon_exit (lttdaemon_is_already_running_p (NULL) == TRUE
			  ? EXIT_SUCCESS : EXIT_FAILURE);
	}
      else
	{
	  if (!lttdaemon_is_already_running_p (NULL))
	    die (EXIT_FAILURE, "there is no lttdaemon running\n");
	}

      /* XXX: Why did I need to do this? */

      if (!ltt_trace_channel_exists_p ())
	{
	  die (EXIT_FAILURE, "missing trace channel: %s: %s\n",
	       ltt_trace_pathname, strerror (errno));
	}

      /* Find pid of running daemon. */

      lttdaemon_pid = get_running_lttdaemon_pid ();

      LOG_MSG ("existing lttdaemon is pid %d\n", lttdaemon_pid);

      if (cliopt_pause)
	{
	  lttdaemon_exit (kill (lttdaemon_pid, SIGUSR1));
	}
      else if (cliopt_resume)
	{
	  lttdaemon_exit (kill (lttdaemon_pid, SIGUSR2));
	}
      else if (cliopt_collect_log)
	{
	  lttdaemon_exit (kill (lttdaemon_pid, SIGTERM));
	}
      else if (cliopt_query_status)
	{
	  if ((rc = access (lttdaemon_statusfile, R_OK | F_OK)) != 0)
	    {
	      LOG_ERR ("cannot access status file: %s: %s\n",
		       lttdaemon_statusfile,
		       strerror (errno));
	    }

	  lttdaemon_exit (rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
	}
      else
	{
	  die (EXIT_FAILURE, "unexpected option!\n");
	}
    }

  lttdaemon_exit (EXIT_SUCCESS); /* not reached */
}

/* }}} */

/* {{{ FTW - because WR-Linux uClibc doesn't have a copy */

#define __set_errno(X) {errno = (X);}
#define L_ftw

/* XXX
 *
 * A local copy of ftw() - the Wind River Linux distribution doesn't
 * currently have a copy in uClibc.
 */

/* File tree walker functions.
   Copyright (C) 1996-2001, 2002, 2003 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@cygnus.com>, 1996.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#define _GNU_SOURCE
#include <features.h>


#if defined (__UCLIBC_HAS_LFS__) && defined L_ftw64
#define L_ftw

/* If Large file support is enabled, transparently remap
 * things to use the 64-bit interfaces */
#if defined _FILE_OFFSET_BITS && _FILE_OFFSET_BITS != 64
#undef _FILE_OFFSET_BITS
#define	_FILE_OFFSET_BITS   64
#endif
#ifndef __USE_LARGEFILE64
# define __USE_LARGEFILE64  1
#endif
#ifndef __USE_FILE_OFFSET64
# define __USE_FILE_OFFSET64  1
#endif

#define FTW_NAME ftw64
#define NFTW_NAME nftw64
#define INO_T ino64_t
#define STAT stat64
#define LSTAT lstat64
#define XSTAT stat64
#define FTW_FUNC_T __ftw64_func_t
#define NFTW_FUNC_T __nftw64_func_t
#else
#define FTW_NAME ftw
#define NFTW_NAME nftw
#define INO_T ino_t
#define STAT stat
#define LSTAT lstat
#define XSTAT stat
#define FTW_FUNC_T __ftw_func_t
#define NFTW_FUNC_T __nftw_func_t
#endif

#ifdef L_ftw

#include <alloca.h>
#include <errno.h>
#include <limits.h>
#include <search.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <assert.h>
#include <dirent.h>

/* We define PATH_MAX if the system does not provide a definition.
   This does not artificially limit any operation.  PATH_MAX is simply
   used as a guesstimate for the expected maximal path length.
   Buffers will be enlarged if necessary.  */
#ifndef PATH_MAX
# define PATH_MAX 1024
#endif

struct dir_data
{
    DIR *stream;
    char *content;
};

struct known_object
{
    dev_t dev;
    INO_T ino;
};

struct ftw_data
{
    /* Array with pointers to open directory streams.  */
    struct dir_data **dirstreams;
    size_t actdir;
    size_t maxdir;

    /* Buffer containing name of currently processed object.  */
    char *dirbuf;
    size_t dirbufsize;

    /* Passed as fourth argument to `nftw' callback.  The `base' member
       tracks the content of the `dirbuf'.  */
    struct FTW ftw;

    /* Flags passed to `nftw' function.  0 for `ftw'.  */
    int flags;

    /* Conversion array for flag values.  It is the identity mapping for
       `nftw' calls, otherwise it maps the values to those known by
       `ftw'.  */
    const int *cvt_arr;

    /* Callback function.  We always use the `nftw' form.  */
    NFTW_FUNC_T func;

    /* Device of starting point.  Needed for FTW_MOUNT.  */
    dev_t dev;

    /* Data structure for keeping fingerprints of already processed
       object.  This is needed when not using FTW_PHYS.  */
    void *known_objects;
};


/* Internally we use the FTW_* constants used for `nftw'.  When invoked
   as `ftw', map each flag to the subset of values used by `ftw'.  */
static const int nftw_arr[] =
{
    FTW_F, FTW_D, FTW_DNR, FTW_NS, FTW_SL, FTW_DP, FTW_SLN
};

static const int ftw_arr[] =
{
    FTW_F, FTW_D, FTW_DNR, FTW_NS, FTW_F, FTW_D, FTW_NS
};

/* Forward declarations of local functions.  */
static int ftw_dir (struct ftw_data *data, struct STAT *st);


static int
object_compare (const void *p1, const void *p2)
{
    /* We don't need a sophisticated and useful comparison.  We are only
       interested in equality.  However, we must be careful not to
       accidentally compare `holes' in the structure.  */
    const struct known_object *kp1 = p1, *kp2 = p2;
    int cmp1;
    cmp1 = (kp1->ino > kp2->ino) - (kp1->ino < kp2->ino);
    if (cmp1 != 0)
	return cmp1;
    return (kp1->dev > kp2->dev) - (kp1->dev < kp2->dev);
}


static inline int
add_object (struct ftw_data *data, struct STAT *st)
{
    struct known_object *newp = malloc (sizeof (struct known_object));
    if (newp == NULL)
	return -1;
    newp->dev = st->st_dev;
    newp->ino = st->st_ino;
    return tsearch (newp, &data->known_objects, object_compare) ? 0 : -1;
}


static inline int
find_object (struct ftw_data *data, struct STAT *st)
{
    struct known_object obj;
    obj.dev = st->st_dev;
    obj.ino = st->st_ino;
    return tfind (&obj, &data->known_objects, object_compare) != NULL;
}


static inline int
__attribute ((always_inline))
open_dir_stream (struct ftw_data *data, struct dir_data *dirp)
{
    int result = 0;

    if (data->dirstreams[data->actdir] != NULL)
    {
	/* Oh, oh.  We must close this stream.  Get all remaining
	   entries and store them as a list in the `content' member of
	   the `struct dir_data' variable.  */
	size_t bufsize = 1024;
	char *buf = malloc (bufsize);

	if (buf == NULL)
	    result = -1;
	else
	{
	    DIR *st = data->dirstreams[data->actdir]->stream;
	    struct dirent *d;
	    size_t actsize = 0;

	    while ((d = readdir (st)) != NULL)
	    {
		size_t this_len = _D_EXACT_NAMLEN (d);
		if (actsize + this_len + 2 >= bufsize)
		{
		    char *newp;
		    bufsize += MAX (1024, 2 * this_len);
		    newp = (char *) realloc (buf, bufsize);
		    if (newp == NULL)
		    {
			/* No more memory.  */
			int save_err = errno;
			free (buf);
			__set_errno (save_err);
			result = -1;
			break;
		    }
		    buf = newp;
		}

		*((char *) mempcpy (buf + actsize, d->d_name, this_len))
		    = '\0';
		actsize += this_len + 1;
	    }

	    /* Terminate the list with an additional NUL byte.  */
	    buf[actsize++] = '\0';

	    /* Shrink the buffer to what we actually need.  */
	    data->dirstreams[data->actdir]->content = realloc (buf, actsize);
	    if (data->dirstreams[data->actdir]->content == NULL)
	    {
		int save_err = errno;
		free (buf);
		__set_errno (save_err);
		result = -1;
	    }
	    else
	    {
		closedir (st);
		data->dirstreams[data->actdir]->stream = NULL;
		data->dirstreams[data->actdir] = NULL;
	    }
	}
    }

    /* Open the new stream.  */
    if (result == 0)
    {
	const char *name = ((data->flags & FTW_CHDIR)
		? data->dirbuf + data->ftw.base: data->dirbuf);
	assert (data->dirstreams[data->actdir] == NULL);

	dirp->stream = opendir (name);
	if (dirp->stream == NULL)
	    result = -1;
	else
	{
	    dirp->content = NULL;
	    data->dirstreams[data->actdir] = dirp;

	    if (++data->actdir == data->maxdir)
		data->actdir = 0;
	}
    }

    return result;
}


static int
process_entry (struct ftw_data *data, struct dir_data *dir, const char *name, size_t namlen)
{
    struct STAT st;
    int result = 0;
    int flag = 0;
    size_t new_buflen;

    if (name[0] == '.' && (name[1] == '\0'
		|| (name[1] == '.' && name[2] == '\0')))
	/* Don't process the "." and ".." entries.  */
	return 0;

    new_buflen = data->ftw.base + namlen + 2;
    if (data->dirbufsize < new_buflen)
    {
	/* Enlarge the buffer.  */
	char *newp;

	data->dirbufsize = 2 * new_buflen;
	newp = (char *) realloc (data->dirbuf, data->dirbufsize);
	if (newp == NULL)
	    return -1;
	data->dirbuf = newp;
    }

    *((char *) mempcpy (data->dirbuf + data->ftw.base, name, namlen)) = '\0';

    if ((data->flags & FTW_CHDIR) == 0)
	name = data->dirbuf;

    if (((data->flags & FTW_PHYS)
		? LSTAT (name, &st)
		: XSTAT (name, &st)) < 0)
    {
	if (errno != EACCES && errno != ENOENT)
	    result = -1;
	else if (!(data->flags & FTW_PHYS)
		&& LSTAT (name, &st) == 0
		&& S_ISLNK (st.st_mode))
	    flag = FTW_SLN;
	else
	    flag = FTW_NS;
    }
    else
    {
	if (S_ISDIR (st.st_mode))
	    flag = FTW_D;
	else if (S_ISLNK (st.st_mode))
	    flag = FTW_SL;
	else
	    flag = FTW_F;
    }

    if (result == 0
	    && (flag == FTW_NS
		|| !(data->flags & FTW_MOUNT) || st.st_dev == data->dev))
    {
	if (flag == FTW_D)
	{
	    if ((data->flags & FTW_PHYS)
		    || (!find_object (data, &st)
			/* Remember the object.  */
			&& (result = add_object (data, &st)) == 0))
	    {
		result = ftw_dir (data, &st);

		if (result == 0 && (data->flags & FTW_CHDIR))
		{
		    /* Change back to the parent directory.  */
		    int done = 0;
		    if (dir->stream != NULL)
			if (fchdir (dirfd (dir->stream)) == 0)
			    done = 1;

		    if (!done)
		    {
			if (data->ftw.base == 1)
			{
			    if (chdir ("/") < 0)
				result = -1;
			}
			else
			    if (chdir ("..") < 0)
				result = -1;
		    }
		}
	    }
	}
	else
	    result = (*data->func) (data->dirbuf, &st, data->cvt_arr[flag],
		    &data->ftw);
    }

    return result;
}

static int
ftw_dir (struct ftw_data *data, struct STAT *st)
{
    struct dir_data dir;
    struct dirent *d;
    int previous_base = data->ftw.base;
    int result;
    char *startp;

    /* Open the stream for this directory.  This might require that
       another stream has to be closed.  */
    result = open_dir_stream (data, &dir);
    if (result != 0)
    {
	if (errno == EACCES)
	    /* We cannot read the directory.  Signal this with a special flag.  */
	    result = (*data->func) (data->dirbuf, st, FTW_DNR, &data->ftw);

	return result;
    }

    /* First, report the directory (if not depth-first).  */
    if (!(data->flags & FTW_DEPTH))
    {
	result = (*data->func) (data->dirbuf, st, FTW_D, &data->ftw);
	if (result != 0)
	    return result;
    }

    /* If necessary, change to this directory.  */
    if (data->flags & FTW_CHDIR)
    {
	if (fchdir (dirfd (dir.stream)) < 0)
	{
	    int save_err = errno;
	    closedir (dir.stream);
	    __set_errno (save_err);

	    if (data->actdir-- == 0)
		data->actdir = data->maxdir - 1;
	    data->dirstreams[data->actdir] = NULL;

	    return -1;
	}
    }

    /* Next, update the `struct FTW' information.  */
    ++data->ftw.level;
    startp = strchr (data->dirbuf, '\0');
    /* There always must be a directory name.  */
    assert (startp != data->dirbuf);
    if (startp[-1] != '/')
	*startp++ = '/';
    data->ftw.base = startp - data->dirbuf;

    while (dir.stream != NULL && (d = readdir (dir.stream)) != NULL)
    {
	result = process_entry (data, &dir, d->d_name, _D_EXACT_NAMLEN (d));
	if (result != 0)
	    break;
    }

    if (dir.stream != NULL)
    {
	/* The stream is still open.  I.e., we did not need more
	   descriptors.  Simply close the stream now.  */
	int save_err = errno;

	assert (dir.content == NULL);

	closedir (dir.stream);
	__set_errno (save_err);

	if (data->actdir-- == 0)
	    data->actdir = data->maxdir - 1;
	data->dirstreams[data->actdir] = NULL;
    }
    else
    {
	int save_err;
	char *runp = dir.content;

	while (result == 0 && *runp != '\0')
	{
	    char *endp = strchr (runp, '\0');

	    result = process_entry (data, &dir, runp, endp - runp);

	    runp = endp + 1;
	}

	save_err = errno;
	free (dir.content);
	__set_errno (save_err);
    }

    /* Prepare the return, revert the `struct FTW' information.  */
    data->dirbuf[data->ftw.base - 1] = '\0';
    --data->ftw.level;
    data->ftw.base = previous_base;

    /* Finally, if we process depth-first report the directory.  */
    if (result == 0 && (data->flags & FTW_DEPTH))
	result = (*data->func) (data->dirbuf, st, FTW_DP, &data->ftw);

    return result;
}


static int
ftw_startup (const char *dir, int is_nftw, void *func, int descriptors, int flags)
{
    struct ftw_data data;
    struct STAT st;
    int result = 0;
    int save_err;
    char *cwd = NULL;
    char *cp;

    /* First make sure the parameters are reasonable.  */
    if (dir==NULL || *dir=='\0') {
	__set_errno (ENOENT);
	return -1;
    }
    if ((strlen(dir)+1) > NAME_MAX) {
	__set_errno(ENAMETOOLONG);
	return -1;
    }

    data.maxdir = descriptors < 1 ? 1 : descriptors;
    data.actdir = 0;
    data.dirstreams = (struct dir_data **) alloca (data.maxdir
	    * sizeof (struct dir_data *));
    memset (data.dirstreams, '\0', data.maxdir * sizeof (struct dir_data *));

    /* PATH_MAX is always defined when we get here.  */
    data.dirbufsize = MAX (2 * strlen (dir), PATH_MAX);
    data.dirbuf = (char *) malloc (data.dirbufsize);
    if (data.dirbuf == NULL)
	return -1;
    cp = stpcpy (data.dirbuf, dir);
    /* Strip trailing slashes.  */
    while (cp > data.dirbuf + 1 && cp[-1] == '/')
	--cp;
    *cp = '\0';

    data.ftw.level = 0;

    /* Find basename.  */
    while (cp > data.dirbuf && cp[-1] != '/')
	--cp;
    data.ftw.base = cp - data.dirbuf;

    data.flags = flags;

    /* This assignment might seem to be strange but it is what we want.
       The trick is that the first three arguments to the `ftw' and
       `nftw' callback functions are equal.  Therefore we can call in
       every case the callback using the format of the `nftw' version
       and get the correct result since the stack layout for a function
       call in C allows this.  */
    data.func = (NFTW_FUNC_T) func;

    /* Since we internally use the complete set of FTW_* values we need
       to reduce the value range before calling a `ftw' callback.  */
    data.cvt_arr = is_nftw ? nftw_arr : ftw_arr;

    /* No object known so far.  */
    data.known_objects = NULL;

    /* Now go to the directory containing the initial file/directory.  */
    if (flags & FTW_CHDIR)
    {
	/* GNU extension ahead.  */
	cwd =  getcwd (NULL, 0);
	if (cwd == NULL)
	    result = -1;
	else if (data.ftw.base > 0)
	{
	    /* Change to the directory the file is in.  In data.dirbuf
	       we have a writable copy of the file name.  Just NUL
	       terminate it for now and change the directory.  */
	    if (data.ftw.base == 1)
		/* I.e., the file is in the root directory.  */
		result = chdir ("/");
	    else
	    {
		char ch = data.dirbuf[data.ftw.base - 1];
		data.dirbuf[data.ftw.base - 1] = '\0';
		result = chdir (data.dirbuf);
		data.dirbuf[data.ftw.base - 1] = ch;
	    }
	}
    }

    /* Get stat info for start directory.  */
    if (result == 0)
    {
	const char *name = ((data.flags & FTW_CHDIR)
		? data.dirbuf + data.ftw.base
		: data.dirbuf);

	if (((flags & FTW_PHYS)
		    ? LSTAT (name, &st)
		    : XSTAT (name, &st)) < 0)
	{
	    if (!(flags & FTW_PHYS)
		    && errno == ENOENT
		    && LSTAT (name, &st) == 0
		    && S_ISLNK (st.st_mode))
		result = (*data.func) (data.dirbuf, &st, data.cvt_arr[FTW_SLN],
			&data.ftw);
	    else
		/* No need to call the callback since we cannot say anything
		   about the object.  */
		result = -1;
	}
	else
	{
	    if (S_ISDIR (st.st_mode))
	    {
		/* Remember the device of the initial directory in case
		   FTW_MOUNT is given.  */
		data.dev = st.st_dev;

		/* We know this directory now.  */
		if (!(flags & FTW_PHYS))
		    result = add_object (&data, &st);

		if (result == 0)
		    result = ftw_dir (&data, &st);
	    }
	    else
	    {
		int flag = S_ISLNK (st.st_mode) ? FTW_SL : FTW_F;

		result = (*data.func) (data.dirbuf, &st, data.cvt_arr[flag],
			&data.ftw);
	    }
	}
    }

    /* Return to the start directory (if necessary).  */
    if (cwd != NULL)
    {
	int save_err1 = errno;
	chdir (cwd);
	free (cwd);
	__set_errno (save_err1);
    }

    /* Free all memory.  */
    save_err = errno;
    tdestroy (data.known_objects, free);
    free (data.dirbuf);
    __set_errno (save_err);

    return result;
}

/* Entry points.  */

int FTW_NAME (const char *path, FTW_FUNC_T func, int descriptors)
{
    return ftw_startup (path, 0, func, descriptors, 0);
}

int NFTW_NAME (const char *path, NFTW_FUNC_T func, int descriptors, int flags)
{
    return ftw_startup (path, 1, func, descriptors, flags);
}
#endif

/* }}} */
